#    Copyright 2012 Nicira, Inc
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures
from lxml import etree
import mock
import os_vif
from os_vif import exception as osv_exception
from os_vif import objects as osv_objects
from os_vif.objects import fields as osv_fields
from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_utils.fixture import uuidsentinel as uuids

from nova import exception
from nova.network import model as network_model
from nova import objects
from nova.pci import utils as pci_utils
from nova import test
from nova.tests import fixtures as nova_fixtures
from nova.tests.fixtures import libvirt as fakelibvirt
from nova.tests.unit.virt import fakelibosinfo
from nova.virt.libvirt import config as vconfig
from nova.virt.libvirt import vif

CONF = cfg.CONF


class LibvirtVifTestCase(test.NoDBTestCase):

    gateway_bridge_4 = network_model.IP(address='101.168.1.1', type='gateway')
    dns_bridge_4 = network_model.IP(address='8.8.8.8', type=None)
    ips_bridge_4 = [network_model.IP(address='101.168.1.9', type=None)]
    subnet_bridge_4 = network_model.Subnet(
        cidr='101.168.1.0/24',
        dns=[dns_bridge_4],
        gateway=gateway_bridge_4,
        routes=None,
        dhcp_server='191.168.1.1')

    gateway_bridge_6 = network_model.IP(address='101:1db9::1', type='gateway')
    subnet_bridge_6 = network_model.Subnet(cidr='101:1db9::/64',
        dns=None,
        gateway=gateway_bridge_6,
        ips=None,
        routes=None)

    network_bridge = network_model.Network(id=uuids.network,
        bridge='br0',
        label=None,
        subnets=[subnet_bridge_4, subnet_bridge_6],
        bridge_interface='eth0',
        vlan=99, mtu=9000)

    vif_bridge = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_bridge,
        type=network_model.VIF_TYPE_BRIDGE,
        devname='tap-xxx-yyy-zzz',
        ovs_interfaceid=None)

    network_bridge_neutron = network_model.Network(id=uuids.network,
        bridge=None,
        label=None,
        subnets=[subnet_bridge_4, subnet_bridge_6],
        bridge_interface='eth0',
        vlan=99)

    vif_bridge_neutron = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_bridge_neutron,
        type=None,
        devname='tap-xxx-yyy-zzz',
        ovs_interfaceid=uuids.ovs)

    network_ovs = network_model.Network(id=uuids.network,
        bridge='br0',
        label=None,
        subnets=[subnet_bridge_4, subnet_bridge_6],
        bridge_interface=None,
        vlan=99, mtu=1000)

    network_ivs = network_model.Network(id=uuids.network,
        bridge='br0',
        label=None,
        subnets=[subnet_bridge_4, subnet_bridge_6],
        bridge_interface=None,
        vlan=99)

    vif_agilio_ovs = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_ovs,
        type=network_model.VIF_TYPE_AGILIO_OVS,
        details={'port_filter': False},
        devname='tap-xxx-yyy-zzz',
        ovs_interfaceid=uuids.ovs)

    vif_agilio_ovs_direct = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_ovs,
        type=network_model.VIF_TYPE_AGILIO_OVS,
        vnic_type=network_model.VNIC_TYPE_DIRECT,
        ovs_interfaceid=uuids.ovs,
        devname='tap-xxx-yyy-zzz',
        profile={'pci_slot': '0000:0a:00.1'})

    vif_agilio_ovs_forwarder = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_ovs,
        type=network_model.VIF_TYPE_AGILIO_OVS,
        vnic_type=network_model.VNIC_TYPE_VIRTIO_FORWARDER,
        profile={'pci_slot': '0000:0a:00.1'},
        details={
            network_model.VIF_DETAILS_VHOSTUSER_MODE: 'client',
            network_model.VIF_DETAILS_VHOSTUSER_SOCKET: '/tmp/usv-xxx-yyy-zzz',
            network_model.VIF_DETAILS_VHOSTUSER_OVS_PLUG: True},
        ovs_interfaceid=uuids.ovs, mtu=1500)

    vif_ovs = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_ovs,
        type=network_model.VIF_TYPE_OVS,
        details={'port_filter': False},
        devname='tap-xxx-yyy-zzz',
        ovs_interfaceid=uuids.ovs)

    vif_ovs_direct = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_ovs,
        vnic_type=network_model.VNIC_TYPE_DIRECT,
        profile={'pci_slot': '0000:0a:00.1'},
        type=network_model.VIF_TYPE_OVS,
        details={'port_filter': False},
        ovs_interfaceid=uuids.ovs)

    vif_ovs_filter_cap = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_ovs,
        type=network_model.VIF_TYPE_OVS,
        details={'port_filter': True},
        devname='tap-xxx-yyy-zzz',
        ovs_interfaceid=uuids.ovs,
        delegate_create=True)

    vif_ovs_legacy = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_ovs,
        type=None,
        devname=None,
        ovs_interfaceid=None)

    vif_ivs = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_ivs,
        type=network_model.VIF_TYPE_IVS,
        devname='tap-xxx-yyy-zzz',
        ovs_interfaceid=uuids.ovs)

    vif_none = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_bridge,
        type=None,
        devname='tap-xxx-yyy-zzz',
        ovs_interfaceid=None)

    network_8021 = network_model.Network(id=uuids.network,
        bridge=None,
        label=None,
        subnets=[subnet_bridge_4,
                 subnet_bridge_6],
        interface='eth0',
        vlan=99)

    vif_8021qbh = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_8021,
        type=network_model.VIF_TYPE_802_QBH,
        vnic_type=network_model.VNIC_TYPE_DIRECT,
        ovs_interfaceid=None,
        details={
            network_model.VIF_DETAILS_PROFILEID:
            'MyPortProfile'},
        profile={'pci_vendor_info': '1137:0043',
                 'pci_slot': '0000:0a:00.1',
                 'physical_network': 'phynet1'})

    vif_hw_veb = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_8021,
        type=network_model.VIF_TYPE_HW_VEB,
        vnic_type=network_model.VNIC_TYPE_DIRECT,
        ovs_interfaceid=None,
        details={
            network_model.VIF_DETAILS_VLAN: 100},
        profile={'pci_vendor_info': '1137:0043',
                 'pci_slot': '0000:0a:00.1',
                 'physical_network': 'phynet1'})

    vif_hw_veb_arq = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_8021,
        type=network_model.VIF_TYPE_HW_VEB,
        vnic_type=network_model.VNIC_TYPE_ACCELERATOR_DIRECT,
        ovs_interfaceid=None,
        details={
            network_model.VIF_DETAILS_VLAN: 100},
        profile={'pci_vendor_info': '1137:0043',
                 'pci_slot': '0000:0a:00.1',
                 'physical_network': 'phynet1'})

    vif_hw_veb_trusted = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_8021,
        type=network_model.VIF_TYPE_HW_VEB,
        vnic_type=network_model.VNIC_TYPE_DIRECT,
        ovs_interfaceid=None,
        details={
            network_model.VIF_DETAILS_VLAN: 100},
        profile={'pci_vendor_info': '1137:0043',
                 'pci_slot': '0000:0a:00.1',
                 'physical_network': 'phynet1',
                 'trusted': 'True'})

    vif_hostdev_physical = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_8021,
        type=network_model.VIF_TYPE_HOSTDEV,
        vnic_type=network_model.VNIC_TYPE_DIRECT_PHYSICAL,
        ovs_interfaceid=None,
        profile={'pci_vendor_info': '1137:0043',
                 'pci_slot': '0000:0a:00.1',
                 'physical_network': 'phynet1'})

    vif_hw_veb_macvtap = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_8021,
        type=network_model.VIF_TYPE_HW_VEB,
        vnic_type=network_model.VNIC_TYPE_MACVTAP,
        ovs_interfaceid=None,
        details={network_model.VIF_DETAILS_VLAN: 100},
        profile={'pci_vendor_info': '1137:0043',
                 'pci_slot': '0000:0a:00.1',
                 'physical_network': 'phynet1'})

    vif_8021qbg = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_8021,
        type=network_model.VIF_TYPE_802_QBG,
        ovs_interfaceid=None,
        qbg_params=network_model.VIF8021QbgParams(
        managerid="xxx-yyy-zzz",
        typeid="aaa-bbb-ccc",
        typeidversion="1",
        instanceid="ddd-eee-fff"))

    network_midonet = network_model.Network(id=uuids.network,
        label=None,
        bridge=None,
        subnets=[subnet_bridge_4],
        interface='eth0')

    network_vrouter = network_model.Network(id=uuids.network,
        label=None,
        bridge=None,
        subnets=[subnet_bridge_4, subnet_bridge_6],
        interface='eth0')

    vif_vrouter = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_vrouter,
        type=network_model.VIF_TYPE_VROUTER,
        devname='tap-xxx-yyy-zzz')

    vif_vrouter_direct = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_vrouter,
        type=network_model.VIF_TYPE_VROUTER,
        vnic_type=network_model.VNIC_TYPE_DIRECT,
        profile={'pci_slot': '0000:0a:00.1'},
        devname='tap-xxx-yyy-zzz')

    vif_vrouter_forwarder = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_vrouter,
        type=network_model.VIF_TYPE_VROUTER,
        vnic_type=network_model.VNIC_TYPE_VIRTIO_FORWARDER,
        profile={'pci_slot': '0000:0a:00.1'},
        details={
            network_model.VIF_DETAILS_VHOSTUSER_MODE: 'server',
            network_model.VIF_DETAILS_VHOSTUSER_SOCKET: '/tmp/usv-xxx-yyy-zzz',
            network_model.VIF_DETAILS_VHOSTUSER_VROUTER_PLUG: True})

    vif_contrail_vrouter = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_vrouter,
        type=network_model.VIF_TYPE_VHOSTUSER,
        details={
            network_model.VIF_DETAILS_VHOSTUSER_MODE: 'server',
            network_model.VIF_DETAILS_VHOSTUSER_SOCKET: '/tmp/usv-xxx-yyy-zzz',
            network_model.VIF_DETAILS_VHOSTUSER_VROUTER_PLUG: True})

    vif_ib_hostdev = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_8021,
        type=network_model.VIF_TYPE_IB_HOSTDEV,
        vnic_type=network_model.VNIC_TYPE_DIRECT,
        ovs_interfaceid=None,
        details={network_model.VIF_DETAILS_VLAN: 100},
        profile={'pci_vendor_info': '1137:0043',
                 'pci_slot': '0000:0a:00.1',
                 'physical_network': 'phynet1'})

    vif_midonet = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_midonet,
        type=network_model.VIF_TYPE_MIDONET,
        devname='tap-xxx-yyy-zzz')

    vif_tap = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        type=network_model.VIF_TYPE_TAP,
        devname='tap-xxx-yyy-zzz')

    vif_iovisor = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_bridge,
        type=network_model.VIF_TYPE_IOVISOR,
        devname='tap-xxx-yyy-zzz',
        ovs_interfaceid=None)

    vif_vhostuser = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_bridge,
        type=network_model.VIF_TYPE_VHOSTUSER,
        details={
            network_model.VIF_DETAILS_VHOSTUSER_MODE: 'client',
            network_model.VIF_DETAILS_VHOSTUSER_SOCKET: '/tmp/vif-xxx-yyy-zzz'
        })

    vif_vhostuser_ovs = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_bridge,
        type=network_model.VIF_TYPE_VHOSTUSER,
        details={
            network_model.VIF_DETAILS_VHOSTUSER_MODE: 'client',
            network_model.VIF_DETAILS_VHOSTUSER_SOCKET: '/tmp/usv-xxx-yyy-zzz',
            network_model.VIF_DETAILS_VHOSTUSER_OVS_PLUG: True},
        ovs_interfaceid=uuids.ovs, mtu=1500)

    vif_vhostuser_no_path = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_bridge,
        type=network_model.VIF_TYPE_VHOSTUSER,
        details={network_model.VIF_DETAILS_VHOSTUSER_MODE: 'client'})

    vif_macvtap_vlan = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_8021,
        type=network_model.VIF_TYPE_MACVTAP,
        details={
            network_model.VIF_DETAILS_VLAN: 1,
            network_model.VIF_DETAILS_PHYS_INTERFACE: 'eth0',
            network_model.VIF_DETAILS_MACVTAP_SOURCE: 'eth0.1',
            network_model.VIF_DETAILS_MACVTAP_MODE: 'vepa'})

    vif_macvtap_flat = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_8021,
        type=network_model.VIF_TYPE_MACVTAP,
        details={
            network_model.VIF_DETAILS_PHYS_INTERFACE: 'eth0',
            network_model.VIF_DETAILS_MACVTAP_SOURCE: 'eth0',
            network_model.VIF_DETAILS_MACVTAP_MODE: 'bridge'})

    vif_macvtap_exception = network_model.VIF(id=uuids.vif,
        address='ca:fe:de:ad:be:ef',
        network=network_8021,
        type=network_model.VIF_TYPE_MACVTAP)

    instance = objects.Instance(id=1,
        uuid='f0000000-0000-0000-0000-000000000001',
        project_id=723)

    flavor_1vcpu = objects.Flavor(vcpus=1, memory=512, root_gb=1)

    flavor_2vcpu = objects.Flavor(vcpus=2, memory=512, root_gb=1)

    bandwidth = {
        'quota:vif_inbound_peak': '200',
        'quota:vif_outbound_peak': '20',
        'quota:vif_inbound_average': '100',
        'quota:vif_outbound_average': '10',
        'quota:vif_inbound_burst': '300',
        'quota:vif_outbound_burst': '30'
    }

    def setup_os_vif_objects(self):
        self.os_vif_network = osv_objects.network.Network(
            id="b82c1929-051e-481d-8110-4669916c7915",
            label="Demo Net",
            subnets=osv_objects.subnet.SubnetList(
                objects=[]),
            mtu=9000)

        self.os_vif_bridge = osv_objects.vif.VIFBridge(
            id="dc065497-3c8d-4f44-8fb4-e1d33c16a536",
            address="22:52:25:62:e2:aa",
            plugin="linux_bridge",
            vif_name="nicdc065497-3c",
            bridge_name="br100",
            has_traffic_filtering=False,
            network=self.os_vif_network)

        self.os_vif_ovs_prof = osv_objects.vif.VIFPortProfileOpenVSwitch(
            interface_id="07bd6cea-fb37-4594-b769-90fc51854ee9",
            profile_id="fishfood",
            create_port=True)

        self.os_vif_repr_prof = osv_objects.vif.VIFPortProfileOVSRepresentor(
            interface_id="07bd6cea-fb37-4594-b769-90fc51854ee9",
            profile_id="fishfood",
            representor_name='nicdc065497-3c',
            representor_address='0000:0a:00.1')

        self.os_vif_agilio_ovs = osv_objects.vif.VIFOpenVSwitch(
            id="dc065497-3c8d-4f44-8fb4-e1d33c16a536",
            address="22:52:25:62:e2:aa",
            plugin="agilio_ovs",
            vif_name="nicdc065497-3c",
            bridge_name="br0",
            port_profile=self.os_vif_ovs_prof,
            network=self.os_vif_network)

        self.os_vif_agilio_forwarder = osv_objects.vif.VIFVHostUser(
            id="dc065497-3c8d-4f44-8fb4-e1d33c16a536",
            address="22:52:25:62:e2:aa",
            plugin="agilio_ovs",
            vif_name="nicdc065497-3c",
            path='/var/run/openvswitch/vhudc065497-3c',
            mode='client',
            port_profile=self.os_vif_repr_prof,
            network=self.os_vif_network)

        self.os_vif_agilio_direct = osv_objects.vif.VIFHostDevice(
            id="dc065497-3c8d-4f44-8fb4-e1d33c16a536",
            address="22:52:25:62:e2:aa",
            plugin="agilio_ovs",
            vif_name="nicdc065497-3c",
            dev_type=osv_fields.VIFHostDeviceDevType.ETHERNET,
            dev_address='0000:0a:00.1',
            port_profile=self.os_vif_repr_prof,
            network=self.os_vif_network)

        self.os_vif_ovs = osv_objects.vif.VIFOpenVSwitch(
            id="dc065497-3c8d-4f44-8fb4-e1d33c16a536",
            address="22:52:25:62:e2:aa",
            unplugin="linux_bridge",
            vif_name="nicdc065497-3c",
            bridge_name="br0",
            port_profile=self.os_vif_ovs_prof,
            network=self.os_vif_network)

        self.os_vif_ovs_hybrid = osv_objects.vif.VIFBridge(
            id="dc065497-3c8d-4f44-8fb4-e1d33c16a536",
            address="22:52:25:62:e2:aa",
            unplugin="linux_bridge",
            vif_name="nicdc065497-3c",
            bridge_name="br0",
            port_profile=self.os_vif_ovs_prof,
            has_traffic_filtering=False,
            network=self.os_vif_network)

        self.os_vif_vhostuser = osv_objects.vif.VIFVHostUser(
            id="dc065497-3c8d-4f44-8fb4-e1d33c16a536",
            address="22:52:25:62:e2:aa",
            plugin="openvswitch",
            vif_name="vhudc065497-3c",
            path='/var/run/openvswitch/vhudc065497-3c',
            mode='client',
            port_profile=self.os_vif_ovs_prof,
            network=self.os_vif_network)

        self.os_vif_hostdevice_ethernet = osv_objects.vif.VIFHostDevice(
            id="dc065497-3c8d-4f44-8fb4-e1d33c16a536",
            address="22:52:25:62:e2:aa",
            plugin="linux_bridge",
            vif_name="nicdc065497-3c",
            dev_type=osv_fields.VIFHostDeviceDevType.ETHERNET,
            dev_address='0000:0a:00.1',
            network=self.os_vif_network)

        self.os_vif_hostdevice_generic = osv_objects.vif.VIFHostDevice(
            id="dc065497-3c8d-4f44-8fb4-e1d33c16a536",
            address="22:52:25:62:e2:aa",
            plugin="linux_bridge",
            vif_name="nicdc065497-3c",
            dev_type=osv_fields.VIFHostDeviceDevType.GENERIC,
            dev_address='0000:0a:00.1',
            network=self.os_vif_network)

        self.os_vif_inst_info = osv_objects.instance_info.InstanceInfo(
            uuid="d5b1090c-9e00-4fa4-9504-4b1494857970",
            name="instance-000004da",
            project_id="2f37d7f6-e51a-4a1f-8b6e-b0917ffc8390")

    def setUp(self):
        super(LibvirtVifTestCase, self).setUp()

        self.libvirt = self.useFixture(
            nova_fixtures.LibvirtFixture(stub_os_vif=False))

        # os_vif.initialize is typically done in nova-compute startup
        os_vif.initialize()
        self.setup_os_vif_objects()

        # multiqueue configuration is host OS specific
        self.mock_uname = self.libvirt.mock_uname
        self.mock_uname.return_value = fakelibvirt.os_uname(
            'Linux', '', '5.10.13-200-generic', '', 'x86_64')

    def _get_node(self, xml):
        doc = etree.fromstring(xml)
        ret = doc.findall('./devices/interface')
        self.assertEqual(len(ret), 1)
        return ret[0]

    def _assertMacEquals(self, node, vif):
        mac = node.find("mac").get("address")
        self.assertEqual(mac, vif['address'])

    def _assertTypeEquals(self, node, type, attr, source, br_want):
        self.assertEqual(node.get("type"), type)
        br_name = node.find(attr).get(source)
        self.assertEqual(br_name, br_want)

    def _assertTypeAndMacEquals(self, node, type, attr, source, vif,
                                br_want=None):
        self._assertTypeEquals(node, type, attr, source, br_want)
        self._assertMacEquals(node, vif)

    def _assertModel(self, xml, model_want=None, driver_want=None):
        node = self._get_node(xml)
        if model_want is None:
            ret = node.findall("model")
            self.assertEqual(len(ret), 0)
        else:
            model = node.find("model").get("type")
            self.assertEqual(model, model_want)
        if driver_want is None:
            ret = node.findall("driver")
            self.assertEqual(len(ret), 0)
        else:
            driver = node.find("driver").get("name")
            self.assertEqual(driver, driver_want)

    def _assertTypeAndPciEquals(self, node, type, vif):
        self.assertEqual(node.get("type"), type)
        self._assertPciEqual(node, vif, type="pci")

    def _assertPciEqual(self, node, vif, type=None):
        address = node.find("source").find("address")
        if type:
            addr_type = address.get("type")
            self.assertEqual(type, addr_type)
        pci_slot = "%(domain)s:%(bus)s:%(slot)s.%(func)s" % {
                     'domain': address.get("domain")[2:],
                     'bus': address.get("bus")[2:],
                     'slot': address.get("slot")[2:],
                     'func': address.get("function")[2:]}

        pci_slot_want = vif['profile']['pci_slot']
        self.assertEqual(pci_slot, pci_slot_want)

    def _assertQueueSizeEquals(self, node, rx_want, tx_want):
        rx_queue_size = node.find("driver").get("rx_queue_size")
        tx_queue_size = node.find("driver").get("tx_queue_size")
        self.assertEqual(rx_queue_size, rx_want)
        self.assertEqual(tx_queue_size, tx_want)

    def _assertXmlEqual(self, expectedXmlstr, actualXmlstr):
        if not isinstance(actualXmlstr, str):
            actualXmlstr = etree.tostring(actualXmlstr, encoding='unicode',
                                          pretty_print=True)
        self.assertXmlEqual(expectedXmlstr, actualXmlstr)

    def _get_conf(self):
        conf = vconfig.LibvirtConfigGuest()
        conf.virt_type = "qemu"
        conf.name = "fake-name"
        conf.uuid = "fake-uuid"
        conf.memory = 100 * 1024
        conf.vcpus = 4
        return conf

    def _get_instance_xml(self, driver, vif, image_meta=None, flavor=None):
        if flavor is None:
            flavor = objects.Flavor(name='m1.small',
                                memory_mb=128,
                                vcpus=1,
                                root_gb=0,
                                ephemeral_gb=0,
                                swap=0,
                                extra_specs=dict(self.bandwidth),
                                deleted_at=None,
                                deleted=0,
                                created_at=None, flavorid=1,
                                is_public=True, vcpu_weight=None,
                                id=2, disabled=False, rxtx_factor=1.0)

        if image_meta is None:
            image_meta = objects.ImageMeta.from_dict({})

        conf = self._get_conf()
        nic = driver.get_config(self.instance, vif, image_meta,
                                flavor, CONF.libvirt.virt_type)
        # TODO(stephenfin): There doesn't appear to be any reason we should do
        # this: just return 'nic.to_xml()' and remove '_get_node'
        conf.add_device(nic)
        return conf.to_xml()

    def _test_virtio_multiqueue(self, vcpus, want_queues):
        self.flags(use_virtio_for_bridges=True,
                   virt_type='kvm',
                   group='libvirt')

        flavor = objects.Flavor(name='m1.small',
                    memory_mb=128,
                    vcpus=vcpus,
                    root_gb=0,
                    ephemeral_gb=0,
                    swap=0,
                    deleted_at=None,
                    deleted=0,
                    created_at=None, flavorid=1,
                    is_public=True, vcpu_weight=None,
                    id=2, disabled=False, rxtx_factor=1.0,
                    extra_specs={})

        d = vif.LibvirtGenericVIFDriver()
        image_meta = objects.ImageMeta.from_dict(
            {'properties': {'hw_vif_model': 'virtio',
                            'hw_vif_multiqueue_enabled': 'true'}})
        xml = self._get_instance_xml(d, self.vif_bridge,
                                     image_meta, flavor)

        node = self._get_node(xml)
        driver = node.find("driver").get("name")
        self.assertEqual(driver, 'vhost')
        queues = node.find("driver").get("queues")
        self.assertEqual(queues, want_queues)

    def test_virtio_multiqueue(self):
        self._test_virtio_multiqueue(4, '4')

    def test_virtio_multiqueue_in_kernel_2(self):
        self.mock_uname.return_value = fakelibvirt.os_uname(
            'Linux', '', '2.6.32-21-generic', '', '')
        self._test_virtio_multiqueue(10, '1')

    def test_virtio_multiqueue_in_kernel_3(self):
        self.mock_uname.return_value = fakelibvirt.os_uname(
            'Linux', '', '3.19.0-47-generic', '', '')
        self._test_virtio_multiqueue(10, '8')

    def test_virtio_multiqueue_in_kernel_4(self):
        self.mock_uname.return_value = fakelibvirt.os_uname(
            'Linux', '', '4.2.0-35-generic', '', '')
        self._test_virtio_multiqueue(10, '10')

    def test_virtio_multiqueue_in_kernel_2_max_queues(self):
        self.mock_uname.return_value = fakelibvirt.os_uname(
            'Linux', '', '2.6.32-21-generic', '', '')
        self.flags(max_queues=2, group='libvirt')
        self._test_virtio_multiqueue(10, '2')

    def test_virtio_multiqueue_in_kernel_3_max_queues(self):
        self.mock_uname.return_value = fakelibvirt.os_uname(
            'Linux', '', '3.19.0-47-generic', '', '')
        self.flags(max_queues=2, group='libvirt')
        self._test_virtio_multiqueue(10, '2')

    def test_virtio_multiqueue_in_kernel_4_max_queues(self):
        self.mock_uname.return_value = fakelibvirt.os_uname(
            'Linux', '', '4.2.0-35-generic', '', '')
        self.flags(max_queues=2, group='libvirt')
        self._test_virtio_multiqueue(10, '2')

    def test_vhostuser_os_vif_multiqueue(self):
        d = vif.LibvirtGenericVIFDriver()
        image_meta = objects.ImageMeta.from_dict(
            {'properties': {'hw_vif_model': 'virtio',
                            'hw_vif_multiqueue_enabled': 'true'}})
        flavor = objects.Flavor(
            id=2,
            name='m1.small',
            memory_mb=128,
            vcpus=4,
            root_gb=0,
            ephemeral_gb=0,
            swap=0,
            deleted_at=None,
            deleted=0,
            created_at=None, flavorid=1,
            is_public=True, vcpu_weight=None,
            disabled=False,
            extra_specs={})
        conf = d.get_base_config(None, 'ca:fe:de:ad:be:ef', image_meta,
                                 flavor, 'kvm', 'normal')
        self.assertEqual(4, conf.vhost_queues)
        self.assertEqual('vhost', conf.driver_name)

        d._set_config_VIFVHostUser(self.instance, self.os_vif_vhostuser, conf)
        self.assertEqual(4, conf.vhost_queues)
        self.assertIsNone(conf.driver_name)

    def _test_virtio_config_queue_sizes(
            self, vnic_type=network_model.VNIC_TYPE_NORMAL):
        self.flags(rx_queue_size=512, group='libvirt')
        self.flags(tx_queue_size=1024, group='libvirt')
        v = vif.LibvirtGenericVIFDriver()
        conf = v.get_base_config(
            None, 'ca:fe:de:ad:be:ef', {},
            objects.Flavor(vcpus=2), 'kvm', vnic_type)
        return v, conf

    def test_virtio_vhost_queue_sizes(self):
        _, conf = self._test_virtio_config_queue_sizes()
        self.assertEqual(512, conf.vhost_rx_queue_size)
        self.assertIsNone(conf.vhost_tx_queue_size)

    def test_virtio_vhost_queue_sizes_vnic_type_direct(self):
        _, conf = self._test_virtio_config_queue_sizes(
            vnic_type=network_model.VNIC_TYPE_DIRECT)
        self.assertIsNone(conf.vhost_rx_queue_size)
        self.assertIsNone(conf.vhost_tx_queue_size)

    def test_virtio_vhost_queue_sizes_vnic_type_direct_physical(self):
        _, conf = self._test_virtio_config_queue_sizes(
            vnic_type=network_model.VNIC_TYPE_DIRECT_PHYSICAL)
        self.assertIsNone(conf.vhost_rx_queue_size)
        self.assertIsNone(conf.vhost_tx_queue_size)

    def test_virtio_vhost_queue_sizes_vnic_type_macvtap(self):
        _, conf = self._test_virtio_config_queue_sizes(
            vnic_type=network_model.VNIC_TYPE_MACVTAP)
        self.assertEqual(512, conf.vhost_rx_queue_size)
        self.assertIsNone(conf.vhost_tx_queue_size)

    def test_virtio_vhost_queue_sizes_vnic_type_virtio_forwarder(self):
        _, conf = self._test_virtio_config_queue_sizes(
            vnic_type=network_model.VNIC_TYPE_VIRTIO_FORWARDER)
        self.assertEqual(512, conf.vhost_rx_queue_size)
        self.assertIsNone(conf.vhost_tx_queue_size)

    def test_virtio_vhostuser_osvif_queue_sizes(self):
        v, conf = self._test_virtio_config_queue_sizes()
        v._set_config_VIFVHostUser(self.instance, self.os_vif_vhostuser, conf)
        self.assertEqual(512, conf.vhost_rx_queue_size)
        self.assertEqual(1024, conf.vhost_tx_queue_size)

    def test_multiple_nics(self):
        conf = self._get_conf()
        # Tests multiple nic configuration and that target_dev is
        # set for each
        nics = [{'net_type': 'bridge',
                 'mac_addr': '00:00:00:00:00:0b',
                 'source_dev': 'b_source_dev',
                 'target_dev': 'b_target_dev'},
                {'net_type': 'ethernet',
                 'mac_addr': '00:00:00:00:00:0e',
                 'source_dev': 'e_source_dev',
                 'target_dev': 'e_target_dev'},
                {'net_type': 'direct',
                 'mac_addr': '00:00:00:00:00:0d',
                 'source_dev': 'd_source_dev',
                 'target_dev': 'd_target_dev'}]

        for nic in nics:
            nic_conf = vconfig.LibvirtConfigGuestInterface()
            nic_conf.net_type = nic['net_type']
            nic_conf.target_dev = nic['target_dev']
            nic_conf.mac_addr = nic['mac_addr']
            nic_conf.source_dev = nic['source_dev']
            conf.add_device(nic_conf)

        xml = conf.to_xml()
        doc = etree.fromstring(xml)
        for nic in nics:
            path = "./devices/interface/[@type='%s']" % nic['net_type']
            node = doc.find(path)
            self.assertEqual(nic['net_type'], node.get("type"))
            self.assertEqual(nic['mac_addr'],
                             node.find("mac").get("address"))
            self.assertEqual(nic['target_dev'],
                             node.find("target").get("dev"))

    def test_model_novirtio(self):
        self.flags(use_virtio_for_bridges=False,
                   virt_type='kvm',
                   group='libvirt')

        d = vif.LibvirtGenericVIFDriver()
        xml = self._get_instance_xml(d, self.vif_bridge)
        self._assertModel(xml)

    def test_model_kvm(self):
        self.flags(use_virtio_for_bridges=True,
                   virt_type='kvm',
                   group='libvirt')

        d = vif.LibvirtGenericVIFDriver()
        xml = self._get_instance_xml(d, self.vif_bridge)
        self._assertModel(xml, network_model.VIF_MODEL_VIRTIO)

    def test_model_parallels(self):
        self.flags(use_virtio_for_bridges=True,
                   virt_type='parallels',
                   group='libvirt')

        d = vif.LibvirtGenericVIFDriver()
        xml = self._get_instance_xml(d, self.vif_bridge)
        self._assertModel(xml, network_model.VIF_MODEL_VIRTIO)

    def test_model_kvm_qemu_parallels_custom(self):
        for virt in ('kvm', 'qemu', 'parallels'):
            self.flags(use_virtio_for_bridges=True,
                       virt_type=virt,
                       group='libvirt')

            d = vif.LibvirtGenericVIFDriver()
            if virt == 'parallels':
                supported = (network_model.VIF_MODEL_RTL8139,
                             network_model.VIF_MODEL_E1000)
            elif virt == 'qemu':
                supported = (network_model.VIF_MODEL_LAN9118,
                             network_model.VIF_MODEL_NE2K_PCI,
                             network_model.VIF_MODEL_PCNET,
                             network_model.VIF_MODEL_RTL8139,
                             network_model.VIF_MODEL_E1000,
                             network_model.VIF_MODEL_E1000E,
                             network_model.VIF_MODEL_SPAPR_VLAN,
                             network_model.VIF_MODEL_VMXNET3)
            else:
                supported = (network_model.VIF_MODEL_NE2K_PCI,
                             network_model.VIF_MODEL_PCNET,
                             network_model.VIF_MODEL_RTL8139,
                             network_model.VIF_MODEL_E1000,
                             network_model.VIF_MODEL_E1000E,
                             network_model.VIF_MODEL_SPAPR_VLAN,
                             network_model.VIF_MODEL_VMXNET3)
            for model in supported:
                image_meta = objects.ImageMeta.from_dict(
                    {'properties': {'hw_vif_model': model}})
                xml = self._get_instance_xml(d, self.vif_bridge,
                                             image_meta)
                self._assertModel(xml, model)

    @mock.patch.object(vif.designer, 'set_vif_guest_frontend_config')
    def test_model_with_osinfo(self, mock_set):
        self.flags(use_virtio_for_bridges=True,
                   virt_type='kvm',
                   group='libvirt')
        self.useFixture(fixtures.MonkeyPatch(
            'nova.virt.osinfo.libosinfo',
            fakelibosinfo))
        d = vif.LibvirtGenericVIFDriver()
        image_meta = {'properties': {'os_name': 'fedora22'}}
        image_meta = objects.ImageMeta.from_dict(image_meta)
        flavor = objects.Flavor(
            id=2,
            name='m1.small',
            memory_mb=128,
            vcpus=4,
            root_gb=0,
            ephemeral_gb=0,
            swap=0,
            deleted_at=None,
            deleted=0,
            created_at=None, flavorid=1,
            is_public=True, vcpu_weight=None,
            disabled=False,
            extra_specs={})
        d.get_base_config(None, 'ca:fe:de:ad:be:ef', image_meta,
                          flavor, 'kvm', 'normal')
        mock_set.assert_called_once_with(mock.ANY, 'ca:fe:de:ad:be:ef',
                                         'virtio', None, None, None)

    @mock.patch.object(vif.designer, 'set_vif_guest_frontend_config',
                       wraps=vif.designer.set_vif_guest_frontend_config)
    def _test_model_sriov(self, vinc_type, mock_set):
        """Direct attach vNICs shouldn't retrieve info from image_meta."""
        self.flags(use_virtio_for_bridges=True,
                   virt_type='kvm',
                   group='libvirt')
        self.useFixture(fixtures.MonkeyPatch(
            'nova.virt.osinfo.libosinfo',
            fakelibosinfo))
        d = vif.LibvirtGenericVIFDriver()
        image_meta = objects.ImageMeta.from_dict(
            {'properties': {'hw_vif_model': 'virtio'}})
        conf = d.get_base_config(None, 'ca:fe:de:ad:be:ef', image_meta,
                                 None, 'kvm', vinc_type)
        mock_set.assert_called_once_with(mock.ANY, 'ca:fe:de:ad:be:ef',
                                         None, None, None, None)
        self.assertIsNone(conf.vhost_queues)
        self.assertIsNone(conf.driver_name)
        self.assertIsNone(conf.model)

    def test_model_sriov_direct(self):
        self._test_model_sriov(network_model.VNIC_TYPE_DIRECT)

    def test_model_accelerator_direct(self):
        self._test_model_sriov(network_model.VNIC_TYPE_ACCELERATOR_DIRECT)

    def test_model_qemu(self):
        vif_objs = [
            self.vif_bridge,
            self.vif_8021qbg,
            self.vif_iovisor,
            self.vif_ovs,
        ]

        self.flags(use_virtio_for_bridges=True,
                   virt_type='qemu',
                   group='libvirt')

        for vif_obj in vif_objs:
            d = vif.LibvirtGenericVIFDriver()

            xml = self._get_instance_xml(d, vif_obj)

            doc = etree.fromstring(xml)

            bandwidth = doc.find('./devices/interface/bandwidth')
            self.assertIsNotNone(bandwidth)

            inbound = bandwidth.find('inbound')
            self.assertEqual(inbound.get("average"),
                             self.bandwidth['quota:vif_inbound_average'])
            self.assertEqual(inbound.get("peak"),
                             self.bandwidth['quota:vif_inbound_peak'])
            self.assertEqual(inbound.get("burst"),
                             self.bandwidth['quota:vif_inbound_burst'])

            outbound = bandwidth.find('outbound')
            self.assertEqual(outbound.get("average"),
                             self.bandwidth['quota:vif_outbound_average'])
            self.assertEqual(outbound.get("peak"),
                             self.bandwidth['quota:vif_outbound_peak'])
            self.assertEqual(outbound.get("burst"),
                             self.bandwidth['quota:vif_outbound_burst'])

            self._assertModel(xml, network_model.VIF_MODEL_VIRTIO, "qemu")

    def test_generic_driver_none(self):
        d = vif.LibvirtGenericVIFDriver()
        self.assertRaises(exception.NovaException,
                          self._get_instance_xml,
                          d,
                          self.vif_none)

    def _check_bridge_driver(self, d, vif, br_want):
        xml = self._get_instance_xml(d, vif)
        node = self._get_node(xml)
        self._assertTypeAndMacEquals(node, "bridge", "source", "bridge",
                                     self.vif_bridge, br_want)

    def test_generic_driver_bridge(self):
        d = vif.LibvirtGenericVIFDriver()
        self._check_bridge_driver(d,
                                  self.vif_bridge,
                                  self.vif_bridge['network']['bridge'])

    def _test_hw_veb_op(self, op, vlan):
        self.libvirt.mock_get_vf_num_by_pci_address.return_value = 1
        pci_utils.get_ifname_by_pci_address.side_effect = ['eth1', 'eth13']
        vlan_id = int(vlan)
        port_state = 'up' if vlan_id > 0 else 'down'
        mac = ('00:00:00:00:00:00' if op.__name__ == 'unplug'
               else self.vif_hw_veb_macvtap['address'])
        calls = {
            'get_ifname':
                [mock.call(self.vif_hw_veb_macvtap['profile']['pci_slot'],
                           pf_interface=True),
                 mock.call(self.vif_hw_veb_macvtap['profile']['pci_slot'])],
            'get_vf_num':
                [mock.call(self.vif_hw_veb_macvtap['profile']['pci_slot'])],
            'set_macaddr': [mock.call('eth13', mac, port_state=port_state)]
        }
        op(self.instance, self.vif_hw_veb_macvtap)
        pci_utils.get_ifname_by_pci_address.assert_has_calls(
            calls['get_ifname'])
        self.libvirt.mock_get_vf_num_by_pci_address.assert_has_calls(
            calls['get_vf_num'])
        self.libvirt.mock_set_device_macaddr.assert_has_calls(
            calls['set_macaddr'])
        self.libvirt.mock_set_device_macaddr_and_vlan.assert_called_once_with(
            'eth1', 1, mock.ANY, vlan_id)

    def test_plug_hw_veb(self):
        d = vif.LibvirtGenericVIFDriver()
        details = self.vif_hw_veb_macvtap['details']
        vlan = details[network_model.VIF_DETAILS_VLAN]
        self._test_hw_veb_op(d.plug, vlan)

    def test_unplug_hw_veb(self):
        d = vif.LibvirtGenericVIFDriver()
        self._test_hw_veb_op(d.unplug, 0)

    @mock.patch('nova.virt.libvirt.vif.set_vf_trusted')
    def test_plug_hw_veb_trusted(self, mset_vf_trusted):
        d = vif.LibvirtGenericVIFDriver()
        d.plug(self.instance, self.vif_hw_veb_trusted)
        mset_vf_trusted.assert_called_once_with('0000:0a:00.1', True)

    @mock.patch('nova.virt.libvirt.vif.set_vf_trusted')
    def test_unplug_hw_veb_trusted(self, mset_vf_trusted):
        d = vif.LibvirtGenericVIFDriver()
        d.unplug(self.instance, self.vif_hw_veb_trusted)
        mset_vf_trusted.assert_called_once_with('0000:0a:00.1', False)

    @mock.patch('nova.privsep.libvirt.unplug_plumgrid_vif',
                side_effect=processutils.ProcessExecutionError)
    def test_unplug_iovisor(self, mock_unplug):
        d = vif.LibvirtGenericVIFDriver()
        d.unplug(self.instance, self.vif_iovisor)

    @mock.patch('nova.privsep.linux_net.device_exists')
    @mock.patch('nova.privsep.libvirt.plug_plumgrid_vif')
    @mock.patch('nova.privsep.linux_net.create_tap_dev')
    def test_plug_iovisor(self, mock_create_tap_dev, mock_plug, device_exists):
        device_exists.return_value = True
        d = vif.LibvirtGenericVIFDriver()
        d.plug(self.instance, self.vif_iovisor)
        mock_plug.assert_has_calls(
            [mock.call('tap-xxx-yyy-zzz',
                       self.vif_iovisor['id'],
                       self.vif_iovisor['address'],
                       self.vif_iovisor['network']['id'],
                       self.instance.project_id)])

    def test_direct_plug_with_port_filter_cap_no_nova_firewall(self):
        d = vif.LibvirtGenericVIFDriver()
        br_want = self.vif_midonet['devname']
        xml = self._get_instance_xml(d, self.vif_ovs_filter_cap)
        node = self._get_node(xml)
        self._assertTypeAndMacEquals(node, "ethernet", "target", "dev",
                                     self.vif_ovs_filter_cap, br_want)

    def test_ib_hostdev_driver(self):
        d = vif.LibvirtGenericVIFDriver()
        xml = self._get_instance_xml(d, self.vif_ib_hostdev)
        doc = etree.fromstring(xml)
        node = doc.findall('./devices/hostdev')[0]
        self.assertEqual(1, len(node))
        self._assertPciEqual(node, self.vif_ib_hostdev)

    def test_midonet_ethernet_vif_driver(self):
        d = vif.LibvirtGenericVIFDriver()
        br_want = self.vif_midonet['devname']
        xml = self._get_instance_xml(d, self.vif_midonet)
        node = self._get_node(xml)
        self._assertTypeAndMacEquals(node, "ethernet", "target", "dev",
                                     self.vif_midonet, br_want)

    def test_tap_ethernet_vif_driver(self):
        d = vif.LibvirtGenericVIFDriver()
        br_want = self.vif_tap['devname']
        xml = self._get_instance_xml(d, self.vif_tap)
        node = self._get_node(xml)
        self._assertTypeAndMacEquals(node, "ethernet", "target", "dev",
                                     self.vif_tap, br_want)

    @mock.patch('nova.privsep.linux_net.device_exists', return_value=True)
    @mock.patch('nova.privsep.linux_net.set_device_mtu')
    @mock.patch('nova.privsep.linux_net.create_tap_dev')
    def test_plug_tap_kvm_virtio(
            self, mock_create_tap_dev, mock_set_mtu, mock_device_exists):

        d1 = vif.LibvirtGenericVIFDriver()
        ins = objects.Instance(
            id=1, uuid='f0000000-0000-0000-0000-000000000001',
            image_ref=uuids.image_ref, flavor=self.flavor_2vcpu,
            project_id=723, system_metadata={}
        )
        d1.plug(ins, self.vif_tap)
        mock_create_tap_dev.assert_called_once_with(
            'tap-xxx-yyy-zzz', None, multiqueue=False)

        mock_create_tap_dev.reset_mock()

        d2 = vif.LibvirtGenericVIFDriver()
        mq_ins = objects.Instance(
            id=1, uuid='f0000000-0000-0000-0000-000000000001',
            image_ref=uuids.image_ref, flavor=self.flavor_2vcpu,
            system_metadata={
                'image_hw_vif_multiqueue_enabled': 'True'
            },
        )
        d2.plug(mq_ins, self.vif_tap)
        mock_create_tap_dev.assert_called_once_with(
            'tap-xxx-yyy-zzz', None, multiqueue=True)

    @mock.patch('nova.privsep.linux_net.device_exists', return_value=True)
    @mock.patch('nova.privsep.linux_net.set_device_mtu')
    @mock.patch('nova.privsep.linux_net.create_tap_dev')
    def test_plug_tap_mq_ignored_1vcpu(
            self, mock_create_tap_dev, mock_set_mtu, mock_device_exists):

        d1 = vif.LibvirtGenericVIFDriver()
        mq_ins = objects.Instance(
            id=1, uuid='f0000000-0000-0000-0000-000000000001',
            image_ref=uuids.image_ref, flavor=self.flavor_1vcpu,
            project_id=723, system_metadata={
                'image_hw_vif_multiqueue_enabled': 'True',
            }
        )
        d1.plug(mq_ins, self.vif_tap)
        mock_create_tap_dev.assert_called_once_with(
            'tap-xxx-yyy-zzz', None, multiqueue=False)

    @mock.patch('nova.privsep.linux_net.device_exists', return_value=True)
    @mock.patch('nova.privsep.linux_net.set_device_mtu')
    @mock.patch('nova.privsep.linux_net.create_tap_dev')
    def test_plug_tap_mq_ignored_virt_type(
            self, mock_create_tap_dev, mock_set_mtu, mock_device_exists):

        self.flags(use_virtio_for_bridges=True,
                   virt_type='lxc',
                   group='libvirt')

        d1 = vif.LibvirtGenericVIFDriver()
        ins = objects.Instance(
            id=1, uuid='f0000000-0000-0000-0000-000000000001',
            image_ref=uuids.image_ref, flavor=self.flavor_2vcpu,
            project_id=723,
            system_metadata={
                'image_hw_vif_multiqueue_enabled': 'True'
            },
        )
        d1.plug(ins, self.vif_tap)
        mock_create_tap_dev.assert_called_once_with(
            'tap-xxx-yyy-zzz', None, multiqueue=False)

    @mock.patch('nova.privsep.linux_net.device_exists', return_value=True)
    @mock.patch('nova.privsep.linux_net.set_device_mtu')
    @mock.patch('nova.privsep.linux_net.create_tap_dev')
    def test_plug_tap_mq_ignored_vif_model(
            self, mock_create_tap_dev, mock_set_mtu, mock_device_exists):

        d1 = vif.LibvirtGenericVIFDriver()
        ins = objects.Instance(
            id=1, uuid='f0000000-0000-0000-0000-000000000001',
            image_ref=uuids.image_ref, flavor=self.flavor_2vcpu,
            project_id=723,
            system_metadata={
                'image_hw_vif_multiqueue_enabled': 'True',
                'image_hw_vif_model': 'e1000',
            },
        )
        d1.plug(ins, self.vif_tap)
        mock_create_tap_dev.assert_called_once_with(
            'tap-xxx-yyy-zzz', None, multiqueue=False)

    def test_unplug_tap(self):
        d = vif.LibvirtGenericVIFDriver()
        d.unplug(self.instance, self.vif_tap)

    def test_generic_8021qbh_driver(self):
        d = vif.LibvirtGenericVIFDriver()
        xml = self._get_instance_xml(d, self.vif_8021qbh)
        node = self._get_node(xml)
        self._assertTypeAndPciEquals(node, "hostdev", self.vif_8021qbh)
        self._assertMacEquals(node, self.vif_8021qbh)
        vp = node.find("virtualport")
        self.assertEqual(vp.get("type"), "802.1Qbh")
        profile_id_found = False
        for p_elem in vp.findall("parameters"):
            details = self.vif_8021qbh["details"]
            profile_id = p_elem.get("profileid", None)
            if profile_id:
                self.assertEqual(profile_id,
                                 details[network_model.VIF_DETAILS_PROFILEID])
                profile_id_found = True

        self.assertTrue(profile_id_found)

    def test_hw_veb_driver(self):
        d = vif.LibvirtGenericVIFDriver()
        xml = self._get_instance_xml(d, self.vif_hw_veb)
        node = self._get_node(xml)
        self._assertTypeAndPciEquals(node, "hostdev", self.vif_hw_veb)
        self._assertMacEquals(node, self.vif_hw_veb)

        conf = vconfig.LibvirtConfigGuestInterface()
        conf.parse_dom(node)
        self.assertEqual(conf.vlan, self.vif_hw_veb["details"]["vlan"])

    def test_hostdev_physical_driver(self):
        d = vif.LibvirtGenericVIFDriver()
        xml = self._get_instance_xml(d, self.vif_hostdev_physical)
        doc = etree.fromstring(xml)
        node = doc.findall('./devices/hostdev')[0]
        self.assertEqual(1, len(node))
        self._assertPciEqual(node, self.vif_hostdev_physical)

    def test_hw_veb_driver_macvtap(self):
        pci_utils.get_ifname_by_pci_address.return_value = 'eth1'
        d = vif.LibvirtGenericVIFDriver()
        xml = self._get_instance_xml(d, self.vif_hw_veb_macvtap)
        node = self._get_node(xml)
        self.assertEqual(node.get("type"), "direct")
        self._assertTypeEquals(node, "direct", "source",
                               "dev", "eth1")
        self._assertTypeEquals(node, "direct", "source",
                               "mode", "passthrough")
        self._assertMacEquals(node, self.vif_hw_veb_macvtap)
        vlan = node.find("vlan").find("tag").get("id")
        vlan_want = self.vif_hw_veb["details"]["vlan"]
        self.assertEqual(int(vlan), vlan_want)

    def test_driver_macvtap_vlan(self):
        d = vif.LibvirtGenericVIFDriver()
        xml = self._get_instance_xml(d, self.vif_macvtap_vlan)
        node = self._get_node(xml)
        self.assertEqual(node.get("type"), "direct")
        self._assertTypeEquals(node, "direct", "source",
                               "dev", "eth0.1")
        self._assertTypeEquals(node, "direct", "source",
                               "mode", "vepa")
        self._assertMacEquals(node, self.vif_macvtap_vlan)

    def test_driver_macvtap_flat(self):
        d = vif.LibvirtGenericVIFDriver()
        xml = self._get_instance_xml(d, self.vif_macvtap_flat)
        node = self._get_node(xml)
        self.assertEqual(node.get("type"), "direct")
        self._assertTypeEquals(node, "direct", "source",
                               "dev", "eth0")
        self._assertTypeEquals(node, "direct", "source",
                               "mode", "bridge")
        self._assertMacEquals(node, self.vif_macvtap_flat)

    def test_driver_macvtap_exception(self):
        d = vif.LibvirtGenericVIFDriver()
        e = self.assertRaises(exception.VifDetailsMissingMacvtapParameters,
                          self._get_instance_xml,
                          d,
                          self.vif_macvtap_exception)
        self.assertIn('macvtap_source', str(e))
        self.assertIn('macvtap_mode', str(e))
        self.assertIn('physical_interface', str(e))

    def test_hw_veb_driver_arq(self):
        d = vif.LibvirtGenericVIFDriver()
        xml = self._get_instance_xml(d, self.vif_hw_veb_arq)
        node = self._get_node(xml)
        self._assertTypeAndPciEquals(node, "hostdev", self.vif_hw_veb_arq)
        self._assertMacEquals(node, self.vif_hw_veb_arq)

        conf = vconfig.LibvirtConfigGuestInterface()
        conf.parse_dom(node)
        self.assertEqual(conf.vlan, self.vif_hw_veb_arq["details"]["vlan"])

    @mock.patch('nova.virt.libvirt.vif.ensure_vlan')
    def test_macvtap_plug_vlan(self, ensure_vlan_mock):
        d = vif.LibvirtGenericVIFDriver()
        d.plug(self.instance, self.vif_macvtap_vlan)
        ensure_vlan_mock.assert_called_once_with(1, 'eth0', interface='eth0.1')

    @mock.patch('nova.virt.libvirt.vif.ensure_vlan')
    def test_macvtap_plug_flat(self, ensure_vlan_mock):
        d = vif.LibvirtGenericVIFDriver()
        d.plug(self.instance, self.vif_macvtap_flat)
        self.assertFalse(ensure_vlan_mock.called)

    def test_generic_iovisor_driver(self):
        d = vif.LibvirtGenericVIFDriver()
        br_want = self.vif_iovisor['devname']
        xml = self._get_instance_xml(d, self.vif_iovisor)
        node = self._get_node(xml)
        self._assertTypeAndMacEquals(node, "ethernet", "target", "dev",
                                     self.vif_iovisor, br_want)

    def test_generic_8021qbg_driver(self):
        d = vif.LibvirtGenericVIFDriver()
        xml = self._get_instance_xml(d, self.vif_8021qbg)

        node = self._get_node(xml)
        self._assertTypeEquals(node, "direct", "source", "dev", "eth0")
        self._assertMacEquals(node, self.vif_8021qbg)

        vp = node.find("virtualport")
        self.assertEqual(vp.get("type"), "802.1Qbg")
        manager_id_found = False
        type_id_found = False
        typeversion_id_found = False
        instance_id_found = False
        for p_elem in vp.findall("parameters"):
            wantparams = self.vif_8021qbg['qbg_params']
            manager_id = p_elem.get("managerid", None)
            type_id = p_elem.get("typeid", None)
            typeversion_id = p_elem.get("typeidversion", None)
            instance_id = p_elem.get("instanceid", None)
            if manager_id:
                self.assertEqual(manager_id,
                                 wantparams['managerid'])
                manager_id_found = True
            if type_id:
                self.assertEqual(type_id,
                                 wantparams['typeid'])
                type_id_found = True
            if typeversion_id:
                self.assertEqual(typeversion_id,
                                 wantparams['typeidversion'])
                typeversion_id_found = True
            if instance_id:
                self.assertEqual(instance_id,
                                 wantparams['instanceid'])
                instance_id_found = True

        self.assertTrue(manager_id_found)
        self.assertTrue(type_id_found)
        self.assertTrue(typeversion_id_found)
        self.assertTrue(instance_id_found)

    def test_vhostuser_driver(self):
        d = vif.LibvirtGenericVIFDriver()
        xml = self._get_instance_xml(d, self.vif_vhostuser)
        node = self._get_node(xml)
        self.assertEqual(node.get("type"),
                         network_model.VIF_TYPE_VHOSTUSER)

        self._assertTypeEquals(node, network_model.VIF_TYPE_VHOSTUSER,
                               "source", "mode", "client")
        self._assertTypeEquals(node, network_model.VIF_TYPE_VHOSTUSER,
                               "source", "path", "/tmp/vif-xxx-yyy-zzz")
        self._assertTypeEquals(node, network_model.VIF_TYPE_VHOSTUSER,
                               "source", "type", "unix")
        self._assertMacEquals(node, self.vif_vhostuser)
        self._assertModel(xml, network_model.VIF_MODEL_VIRTIO)

    def test_vhostuser_driver_queue_sizes(self):
        self.flags(rx_queue_size=512, group='libvirt')
        self.flags(tx_queue_size=1024, group='libvirt')
        d = vif.LibvirtGenericVIFDriver()
        xml = self._get_instance_xml(d, self.vif_vhostuser)
        node = self._get_node(xml)
        self._assertQueueSizeEquals(node, "512", "1024")

    def test_vhostuser_driver_no_path(self):
        d = vif.LibvirtGenericVIFDriver()

        self.assertRaises(exception.VifDetailsMissingVhostuserSockPath,
                          self._get_instance_xml,
                          d,
                          self.vif_vhostuser_no_path)

    def test_vhostuser_driver_ovs(self):
        d = vif.LibvirtGenericVIFDriver()
        xml = self._get_instance_xml(d,
                                     self.vif_vhostuser_ovs)
        node = self._get_node(xml)
        self.assertEqual(node.get("type"),
                         network_model.VIF_TYPE_VHOSTUSER)

        self._assertTypeEquals(node, network_model.VIF_TYPE_VHOSTUSER,
                               "source", "mode", "client")
        self._assertTypeEquals(node, network_model.VIF_TYPE_VHOSTUSER,
                               "source", "path", "/tmp/usv-xxx-yyy-zzz")
        self._assertTypeEquals(node, network_model.VIF_TYPE_VHOSTUSER,
                               "source", "type", "unix")
        self._assertMacEquals(node, self.vif_vhostuser_ovs)
        self._assertModel(xml, network_model.VIF_MODEL_VIRTIO)

    def test_ovs_direct(self):
        d = vif.LibvirtGenericVIFDriver()
        xml = self._get_instance_xml(d, self.vif_ovs_direct)
        node = self._get_node(xml)
        self._assertTypeAndPciEquals(node,
                                     "hostdev",
                                     self.vif_ovs_direct)
        self._assertMacEquals(node, self.vif_ovs_direct)

    def test_agilio_ovs_direct(self):
        d = vif.LibvirtGenericVIFDriver()
        xml = self._get_instance_xml(d, self.vif_agilio_ovs_direct)
        node = self._get_node(xml)
        self._assertTypeAndPciEquals(node,
                                     "hostdev",
                                     self.vif_agilio_ovs_direct)
        self._assertMacEquals(node, self.vif_agilio_ovs_direct)

    def test_agilio_ovs_forwarder(self):
        d = vif.LibvirtGenericVIFDriver()
        xml = self._get_instance_xml(d,
                                     self.vif_agilio_ovs_forwarder)
        node = self._get_node(xml)
        self.assertEqual(node.get("type"),
                         network_model.VIF_TYPE_VHOSTUSER)

        self._assertTypeEquals(node, network_model.VIF_TYPE_VHOSTUSER,
                               "source", "mode", "client")
        self._assertTypeEquals(node, network_model.VIF_TYPE_VHOSTUSER,
                               "source", "path", "/tmp/usv-xxx-yyy-zzz")
        self._assertTypeEquals(node, network_model.VIF_TYPE_VHOSTUSER,
                               "source", "type", "unix")
        self._assertMacEquals(node, self.vif_agilio_ovs_forwarder)
        self._assertModel(xml, network_model.VIF_MODEL_VIRTIO)

    def test_ivs_ethernet_driver(self):
        d = vif.LibvirtGenericVIFDriver()
        xml = self._get_instance_xml(d, self.vif_ivs)
        node = self._get_node(xml)
        dev_want = self.vif_ivs['devname']
        self._assertTypeAndMacEquals(node, "ethernet", "target", "dev",
                                     self.vif_ivs, dev_want)
        script = node.find("script")
        self.assertIsNone(script)

    def test_vrouter(self):
        """Test for the Contrail / Tungsten Fabric kernel datapath."""
        d = vif.LibvirtGenericVIFDriver()
        dev_want = self.vif_vrouter['devname']
        xml = self._get_instance_xml(d, self.vif_vrouter)
        node = self._get_node(xml)
        self._assertTypeAndMacEquals(node, "ethernet", "target", "dev",
                                     self.vif_vrouter, dev_want)

    def test_vrouter_direct(self):
        """Test for Contrail / Tungsten Fabric direct offloaded datapath."""
        d = vif.LibvirtGenericVIFDriver()
        xml = self._get_instance_xml(d, self.vif_vrouter_direct)
        node = self._get_node(xml)
        self._assertTypeAndPciEquals(node,
                                     "hostdev",
                                     self.vif_vrouter_direct)
        self._assertMacEquals(node, self.vif_vrouter_direct)

    def test_vrouter_forwarder(self):
        """Test for Contrail / Tungsten Fabric indirect offloaded datapath."""
        d = vif.LibvirtGenericVIFDriver()
        xml = self._get_instance_xml(d,
                                     self.vif_vrouter_forwarder)
        node = self._get_node(xml)
        self.assertEqual(node.get("type"),
                         network_model.VIF_TYPE_VHOSTUSER)

        self._assertTypeEquals(node, network_model.VIF_TYPE_VHOSTUSER,
                               "source", "mode", "server")
        self._assertTypeEquals(node, network_model.VIF_TYPE_VHOSTUSER,
                               "source", "path", "/tmp/usv-xxx-yyy-zzz")
        self._assertTypeEquals(node, network_model.VIF_TYPE_VHOSTUSER,
                               "source", "type", "unix")
        self._assertMacEquals(node, self.vif_vrouter_forwarder)
        self._assertModel(xml, network_model.VIF_MODEL_VIRTIO)

    def test_contrail_vrouter(self):
        """Test for the Contrail / Tungsten Fabric DPDK datapath."""
        d = vif.LibvirtGenericVIFDriver()
        xml = self._get_instance_xml(d,
                                     self.vif_contrail_vrouter)
        node = self._get_node(xml)
        self.assertEqual(node.get("type"),
                         network_model.VIF_TYPE_VHOSTUSER)

        self._assertTypeEquals(node, network_model.VIF_TYPE_VHOSTUSER,
                               "source", "mode", "server")
        self._assertTypeEquals(node, network_model.VIF_TYPE_VHOSTUSER,
                               "source", "path", "/tmp/usv-xxx-yyy-zzz")
        self._assertTypeEquals(node, network_model.VIF_TYPE_VHOSTUSER,
                               "source", "type", "unix")
        self._assertMacEquals(node, self.vif_contrail_vrouter)
        self._assertModel(xml, network_model.VIF_MODEL_VIRTIO)

    @mock.patch("nova.network.os_vif_util.nova_to_osvif_instance")
    @mock.patch("nova.network.os_vif_util.nova_to_osvif_vif")
    @mock.patch.object(os_vif, "plug")
    def _test_osvif_plug(self, fail, mock_plug,
                         mock_convert_vif, mock_convert_inst):
        mock_convert_vif.return_value = self.os_vif_bridge
        mock_convert_inst.return_value = self.os_vif_inst_info

        d = vif.LibvirtGenericVIFDriver()
        if fail:
            mock_plug.side_effect = osv_exception.ExceptionBase("Wibble")
            self.assertRaises(exception.NovaException,
                              d.plug,
                              self.instance, self.vif_bridge)
        else:
            d.plug(self.instance, self.vif_bridge)

        mock_plug.assert_called_once_with(self.os_vif_bridge,
                                          self.os_vif_inst_info)

    def test_osvif_plug_normal(self):
        self._test_osvif_plug(False)

    def test_osvif_plug_fail(self):
        self._test_osvif_plug(True)

    @mock.patch("nova.network.os_vif_util.nova_to_osvif_instance")
    @mock.patch("nova.network.os_vif_util.nova_to_osvif_vif")
    @mock.patch.object(os_vif, "unplug")
    def _test_osvif_unplug(self, fail, mock_unplug,
                         mock_convert_vif, mock_convert_inst):
        mock_convert_vif.return_value = self.os_vif_bridge
        mock_convert_inst.return_value = self.os_vif_inst_info

        d = vif.LibvirtGenericVIFDriver()
        if fail:
            mock_unplug.side_effect = osv_exception.ExceptionBase("Wibble")
            self.assertRaises(exception.NovaException,
                              d.unplug,
                              self.instance, self.vif_bridge)
        else:
            d.unplug(self.instance, self.vif_bridge)

        mock_unplug.assert_called_once_with(self.os_vif_bridge,
                                            self.os_vif_inst_info)

    def test_osvif_unplug_normal(self):
        self._test_osvif_unplug(False)

    def test_osvif_unplug_fail(self):
        self._test_osvif_unplug(True)

    @mock.patch("nova.network.os_vif_util.nova_to_osvif_instance")
    @mock.patch("nova.network.os_vif_util.nova_to_osvif_vif")
    def _test_config_os_vif(self, os_vif_model, vif_model,
            expected_xml, mock_convert_vif,
            mock_convert_inst):
        mock_convert_vif.return_value = os_vif_model
        mock_convert_inst.return_value = self.os_vif_inst_info

        d = vif.LibvirtGenericVIFDriver()
        xml = self._get_instance_xml(d, vif_model)
        node = self._get_node(xml)
        node_xml = etree.tostring(node).decode()
        self._assertXmlEqual(expected_xml, node_xml)

    def test_config_os_vif_bridge(self):
        os_vif_type = self.os_vif_bridge
        vif_type = self.vif_bridge

        expected_xml = """
            <interface type="bridge">
             <mac address="22:52:25:62:e2:aa"/>
             <model type="virtio"/>
             <source bridge="br100"/>
             <mtu size="9000"/>
             <target dev="nicdc065497-3c"/>
             <bandwidth>
              <inbound average="100" peak="200" burst="300"/>
              <outbound average="10" peak="20" burst="30"/>
             </bandwidth>
            </interface>"""

        self._test_config_os_vif(os_vif_type, vif_type, expected_xml)

    def test_config_os_vif_bridge_nofw(self):
        os_vif_type = self.os_vif_bridge
        vif_type = self.vif_bridge

        expected_xml = """
            <interface type="bridge">
             <mac address="22:52:25:62:e2:aa"/>
             <model type="virtio"/>
             <source bridge="br100"/>
             <mtu size="9000"/>
             <target dev="nicdc065497-3c"/>
             <bandwidth>
              <inbound average="100" peak="200" burst="300"/>
              <outbound average="10" peak="20" burst="30"/>
             </bandwidth>
            </interface>"""

        self._test_config_os_vif(os_vif_type, vif_type, expected_xml)

    def test_config_os_vif_vhostuser(self):
        os_vif_type = self.os_vif_vhostuser
        vif_type = self.vif_vhostuser

        expected_xml = """
            <interface type="vhostuser">
             <mac address="22:52:25:62:e2:aa"/>
             <model type="virtio"/>
             <source mode="client"
              path="/var/run/openvswitch/vhudc065497-3c" type="unix"/>
             <target dev="vhudc065497-3c"/>
            </interface>"""

        self._test_config_os_vif(os_vif_type, vif_type, expected_xml)

    def test_config_os_vif_agilio_ovs_fallthrough(self):
        os_vif_type = self.os_vif_agilio_ovs
        vif_type = self.vif_agilio_ovs

        expected_xml = """
            <interface type="ethernet">
             <mac address="22:52:25:62:e2:aa"/>
             <model type="virtio"/>
             <mtu size="9000"/>
             <target dev="nicdc065497-3c"/>
             <bandwidth>
              <inbound average="100" peak="200" burst="300"/>
              <outbound average="10" peak="20" burst="30"/>
             </bandwidth>
            </interface>"""

        self._test_config_os_vif(os_vif_type, vif_type, expected_xml)

    def test_config_os_vif_agilio_ovs_forwarder(self):
        os_vif_type = self.os_vif_agilio_forwarder
        vif_type = self.vif_agilio_ovs_forwarder

        expected_xml = """
            <interface type="vhostuser">
             <mac address="22:52:25:62:e2:aa"/>
             <model type="virtio"/>
             <source mode="client"
              path="/var/run/openvswitch/vhudc065497-3c" type="unix"/>
             <target dev="nicdc065497-3c"/>
            </interface>"""

        self._test_config_os_vif(os_vif_type, vif_type, expected_xml)

    def test_config_os_vif_agilio_ovs_direct(self):
        os_vif_type = self.os_vif_agilio_direct
        vif_type = self.vif_agilio_ovs_direct

        expected_xml = """
            <interface type="hostdev" managed="yes">
             <mac address="22:52:25:62:e2:aa"/>
             <source>
              <address type="pci" domain="0x0000"
               bus="0x0a" slot="0x00" function="0x1"/>
             </source>
            </interface>"""

        self._test_config_os_vif(os_vif_type, vif_type, expected_xml)

    def test_config_os_vif_ovs(self):
        os_vif_type = self.os_vif_ovs
        vif_type = self.vif_ovs

        expected_xml = """
            <interface type="ethernet">
             <mac address="22:52:25:62:e2:aa"/>
             <model type="virtio"/>
             <mtu size="9000"/>
             <target dev="nicdc065497-3c"/>
             <bandwidth>
              <inbound average="100" peak="200" burst="300"/>
              <outbound average="10" peak="20" burst="30"/>
             </bandwidth>
            </interface>"""

        self._test_config_os_vif(os_vif_type, vif_type, expected_xml)

    def test_config_os_vif_ovs_hybrid(self):
        os_vif_type = self.os_vif_ovs_hybrid
        vif_type = self.vif_ovs

        expected_xml = """
            <interface type="bridge">
             <mac address="22:52:25:62:e2:aa"/>
             <model type="virtio"/>
             <source bridge="br0"/>
             <mtu size="9000"/>
             <target dev="nicdc065497-3c"/>
             <bandwidth>
              <inbound average="100" peak="200" burst="300"/>
              <outbound average="10" peak="20" burst="30"/>
             </bandwidth>
            </interface>"""

        self._test_config_os_vif(os_vif_type, vif_type, expected_xml)

    def test_config_os_vif_hostdevice_ethernet(self):
        os_vif_type = self.os_vif_hostdevice_ethernet
        vif_type = self.vif_bridge

        expected_xml = """
            <interface type="hostdev" managed="yes">
             <mac address="22:52:25:62:e2:aa"/>
             <source>
              <address type="pci" domain="0x0000"
               bus="0x0a" slot="0x00" function="0x1"/>
             </source>
            </interface>"""

        self._test_config_os_vif(os_vif_type, vif_type, expected_xml)

    @mock.patch("nova.network.os_vif_util.nova_to_osvif_instance")
    @mock.patch("nova.network.os_vif_util.nova_to_osvif_vif")
    def test_config_os_vif_hostdevice_generic(self, mock_convert_vif,
                                              mock_convert_inst):
        mock_convert_vif.return_value = self.os_vif_hostdevice_generic
        mock_convert_inst.return_value = self.os_vif_inst_info

        d = vif.LibvirtGenericVIFDriver()

        self.assertRaises(exception.InternalError,
                          self._get_instance_xml, d, self.vif_bridge)



# Copyright 2016 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from oslo_policy import policy as oslo_policy

from nova import exception
from nova import policy
from nova import test
from nova.tests import fixtures as nova_fixtures
from nova.tests.functional.api import client
from nova.tests.functional import fixtures as func_fixtures
from nova.tests.functional import integrated_helpers


class InstanceActionsTestV2(integrated_helpers._IntegratedTestBase):
    """Tests Instance Actions API"""

    def test_get_instance_actions(self):
        server = self._create_server()
        actions = self.api.get_instance_actions(server['id'])
        self.assertEqual('create', actions[0]['action'])

    def test_get_instance_actions_deleted(self):
        server = self._create_server()
        self._delete_server(server)
        self.assertRaises(client.OpenStackApiNotFoundException,
                          self.api.get_instance_actions,
                          server['id'])


class InstanceActionsTestV21(InstanceActionsTestV2):
    api_major_version = 'v2.1'


class InstanceActionsTestV221(InstanceActionsTestV21):
    microversion = '2.21'

    def setUp(self):
        super(InstanceActionsTestV221, self).setUp()
        self.api.microversion = self.microversion

    def test_get_instance_actions_deleted(self):
        server = self._create_server()
        self._delete_server(server)
        actions = self.api.get_instance_actions(server['id'])
        self.assertEqual('delete', actions[0]['action'])
        self.assertEqual('create', actions[1]['action'])


class HypervisorError(Exception):
    """This is just used to make sure the exception type is in the events."""
    pass


class InstanceActionEventFaultsTestCase(
    test.TestCase, integrated_helpers.InstanceHelperMixin):
    """Tests for the instance action event details reporting from the API"""

    def setUp(self):
        super(InstanceActionEventFaultsTestCase, self).setUp()
        # Setup the standard fixtures.
        self.useFixture(nova_fixtures.GlanceFixture(self))
        self.useFixture(nova_fixtures.NeutronFixture(self))
        self.useFixture(func_fixtures.PlacementFixture())
        self.useFixture(nova_fixtures.RealPolicyFixture())

        # Start the compute services.
        self.start_service('conductor')
        self.start_service('scheduler')
        self.compute = self.start_service('compute')
        api_fixture = self.useFixture(nova_fixtures.OSAPIFixture(
            api_version='v2.1'))
        self.api = api_fixture.api
        self.admin_api = api_fixture.admin_api

    def _set_policy_rules(self, overwrite=True):
        rules = {'os_compute_api:os-instance-actions:show': '',
                 'os_compute_api:os-instance-actions:events:details':
                     'project_id:%(project_id)s'}
        policy.set_rules(oslo_policy.Rules.from_dict(rules),
                         overwrite=overwrite)

    def test_instance_action_event_details_non_nova_exception(self):
        """Creates a server using the non-admin user, then reboot it which
        will generate a non-NovaException fault and put the instance into
        ERROR status. Then checks that fault details are visible.
        """

        # Create the server with the non-admin user.
        server = self._build_server(
            networks=[{'port': nova_fixtures.NeutronFixture.port_1['id']}])
        server = self.api.post_server({'server': server})
        server = self._wait_for_state_change(server, 'ACTIVE')

        # Stop the server before rebooting it so that after the driver.reboot
        # method raises an exception, the fake driver does not report the
        # instance power state as running - that will make the compute manager
        # set the instance vm_state to error.
        self.api.post_server_action(server['id'], {'os-stop': None})
        server = self._wait_for_state_change(server, 'SHUTOFF')

        # Stub out the compute driver reboot method to raise a non-nova
        # exception to simulate some error from the underlying hypervisor
        # which in this case we are going to say has sensitive content.
        error_msg = 'sensitive info'
        with mock.patch.object(
                self.compute.manager.driver, 'reboot',
                side_effect=HypervisorError(error_msg)) as mock_reboot:
            reboot_request = {'reboot': {'type': 'HARD'}}
            self.api.post_server_action(server['id'], reboot_request)
            # In this case we wait for the status to change to ERROR using
            # the non-admin user so we can assert the fault details. We also
            # wait for the task_state to be None since the wrap_instance_fault
            # decorator runs before the reverts_task_state decorator so we will
            # be sure the fault is set on the server.
            server = self._wait_for_server_parameter(
                server, {'status': 'ERROR', 'OS-EXT-STS:task_state': None},
                api=self.api)
            mock_reboot.assert_called_once()

        self._set_policy_rules(overwrite=False)

        server_id = server['id']
        # Calls GET on the server actions and verifies that the reboot
        # action expected in the response.
        response = self.api.api_get('/servers/%s/os-instance-actions' %
                                    server_id)
        server_actions = response.body['instanceActions']
        for actions in server_actions:
            if actions['action'] == 'reboot':
                reboot_request_id = actions['request_id']
        # non admin shows instance actions details and verifies the 'details'
        # in the action events via 'request_id', since microversion 2.51 that
        # we can show events, but in microversion 2.84 that we can show
        # 'details' for non-admin.
        self.api.microversion = '2.84'
        action_events_response = self.api.api_get(
            '/servers/%s/os-instance-actions/%s' % (server_id,
                                                    reboot_request_id))
        reboot_action = action_events_response.body['instanceAction']
        # Since reboot action failed, the 'message' property in reboot action
        # should be 'Error', otherwise it's None.
        self.assertEqual('Error', reboot_action['message'])
        reboot_action_events = reboot_action['events']
        # The instance action events from the non-admin user API response
        # should not have 'traceback' in it.
        self.assertNotIn('traceback', reboot_action_events[0])
        # And the sensitive details from the non-nova exception should not be
        # in the details.
        self.assertIn('details', reboot_action_events[0])
        self.assertNotIn(error_msg, reboot_action_events[0]['details'])
        # The exception type class name should be in the details.
        self.assertIn('HypervisorError', reboot_action_events[0]['details'])

        # Get the server fault details for the admin user.
        self.admin_api.microversion = '2.84'
        action_events_response = self.admin_api.api_get(
            '/servers/%s/os-instance-actions/%s' % (server_id,
                                                    reboot_request_id))
        reboot_action = action_events_response.body['instanceAction']
        self.assertEqual('Error', reboot_action['message'])
        reboot_action_events = reboot_action['events']
        # The admin can see the fault details which includes the traceback,
        # and make sure the traceback is there by looking for part of it.
        self.assertIn('traceback', reboot_action_events[0])
        self.assertIn('in reboot_instance',
                      reboot_action_events[0]['traceback'])
        # The exception type class name should be in the details for the admin
        # user as well since the fault handling code cannot distinguish who
        # is going to see the message so it only sets class name.
        self.assertIn('HypervisorError', reboot_action_events[0]['details'])

    def test_instance_action_event_details_with_nova_exception(self):
        """Creates a server using the non-admin user, then reboot it which
        will generate a nova exception fault and put the instance into
        ERROR status. Then checks that fault details are visible.
        """

        # Create the server with the non-admin user.
        server = self._build_server(
            networks=[{'port': nova_fixtures.NeutronFixture.port_1['id']}])
        server = self.api.post_server({'server': server})
        server = self._wait_for_state_change(server, 'ACTIVE')

        # Stop the server before rebooting it so that after the driver.reboot
        # method raises an exception, the fake driver does not report the
        # instance power state as running - that will make the compute manager
        # set the instance vm_state to error.
        self.api.post_server_action(server['id'], {'os-stop': None})
        server = self._wait_for_state_change(server, 'SHUTOFF')

        # Stub out the compute driver reboot method to raise a nova
        # exception 'InstanceRebootFailure' to simulate some error.
        exc_reason = 'reboot failure'
        with mock.patch.object(
                self.compute.manager.driver, 'reboot',
                side_effect=exception.InstanceRebootFailure(reason=exc_reason)
            ) as mock_reboot:
            reboot_request = {'reboot': {'type': 'HARD'}}
            self.api.post_server_action(server['id'], reboot_request)
            # In this case we wait for the status to change to ERROR using
            # the non-admin user so we can assert the fault details. We also
            # wait for the task_state to be None since the wrap_instance_fault
            # decorator runs before the reverts_task_state decorator so we will
            # be sure the fault is set on the server.
            server = self._wait_for_server_parameter(
                server, {'status': 'ERROR', 'OS-EXT-STS:task_state': None},
                api=self.api)
            mock_reboot.assert_called_once()

        self._set_policy_rules(overwrite=False)

        server_id = server['id']
        # Calls GET on the server actions and verifies that the reboot
        # action expected in the response.
        response = self.api.api_get('/servers/%s/os-instance-actions' %
                                    server_id)
        server_actions = response.body['instanceActions']
        for actions in server_actions:
            if actions['action'] == 'reboot':
                reboot_request_id = actions['request_id']

        # non admin shows instance actions details and verifies the 'details'
        # in the action events via 'request_id', since microversion 2.51 that
        # we can show events, but in microversion 2.84 that we can show
        # 'details' for non-admin.
        self.api.microversion = '2.84'
        action_events_response = self.api.api_get(
            '/servers/%s/os-instance-actions/%s' % (server_id,
                                                    reboot_request_id))
        reboot_action = action_events_response.body['instanceAction']
        # Since reboot action failed, the 'message' property in reboot action
        # should be 'Error', otherwise it's None.
        self.assertEqual('Error', reboot_action['message'])
        reboot_action_events = reboot_action['events']
        # The instance action events from the non-admin user API response
        # should not have 'traceback' in it.
        self.assertNotIn('traceback', reboot_action_events[0])
        # The nova exception format message should be in the details.
        self.assertIn('details', reboot_action_events[0])
        self.assertIn(exc_reason, reboot_action_events[0]['details'])

        # Get the server fault details for the admin user.
        self.admin_api.microversion = '2.84'
        action_events_response = self.admin_api.api_get(
            '/servers/%s/os-instance-actions/%s' % (server_id,
                                                    reboot_request_id))
        reboot_action = action_events_response.body['instanceAction']
        self.assertEqual('Error', reboot_action['message'])
        reboot_action_events = reboot_action['events']
        # The admin can see the fault details which includes the traceback,
        # and make sure the traceback is there by looking for part of it.
        self.assertIn('traceback', reboot_action_events[0])
        self.assertIn('in reboot_instance',
                      reboot_action_events[0]['traceback'])
        # The nova exception format message should be in the details.
        self.assertIn(exc_reason, reboot_action_events[0]['details'])



# Copyright 2017 IBM Corp.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from keystoneauth1.adapter import Adapter
from keystoneauth1 import exceptions as kse
import webob

from nova.api.openstack import identity
from nova import test
from nova.tests.unit import fake_requests


class IdentityValidationTest(test.NoDBTestCase):
    """Unit tests for our validation of keystone projects.

    There are times when Nova stores keystone project_id and user_id
    in our database as strings. Until the Pike release none of this
    data was validated, so it was very easy for adminstrators to think
    they were adjusting quota for a project (by name) when instead
    they were just inserting keys in a database that would not get used.

    This is only tested in unit tests through mocking out keystoneauth
    responses because a functional test would need a real keystone or
    keystone simulator.

    The functional code works by using the existing keystone
    credentials and trying to make a /v3/projects/{id} get call. It
    will return a 403 if the user doesn't have enough permissions to
    ask about other projects, a 404 if it does and that project does
    not exist.

    """

    def setUp(self):
        super(IdentityValidationTest, self).setUp()
        get_adap_p = mock.patch('nova.utils.get_ksa_adapter')
        self.addCleanup(get_adap_p.stop)
        self.mock_get_adap = get_adap_p.start()
        self.mock_adap = mock.create_autospec(Adapter)
        self.mock_get_adap.return_value = self.mock_adap

    def validate_common(self):
        self.mock_get_adap.assert_called_once_with(
            'identity', ksa_auth=mock.ANY,
            min_version=(3, 0), max_version=(3, 'latest'))
        self.mock_adap.get.assert_called_once_with('/projects/foo')

    def test_good_id(self):
        """Test response 200.

        This indicates we have permissions, and we have definitively
        found the project exists.

        """
        self.mock_adap.get.return_value = fake_requests.FakeResponse(200)
        self.assertTrue(identity.verify_project_id(mock.MagicMock(), "foo"))
        self.validate_common()

    def test_no_project(self):
        """Test response 404.

        This indicates that we have permissions, and we have
        definitively found the project does not exist.

        """
        self.mock_adap.get.return_value = fake_requests.FakeResponse(404)
        self.assertRaises(webob.exc.HTTPBadRequest,
                          identity.verify_project_id,
                          mock.MagicMock(), "foo")
        self.validate_common()

    def test_unknown_id(self):
        """Test response 403.

        This indicates we don't have permissions. We fail open here
        and assume the project exists.

        """
        self.mock_adap.get.return_value = fake_requests.FakeResponse(403)
        self.assertTrue(identity.verify_project_id(mock.MagicMock(), "foo"))
        self.validate_common()

    def test_unknown_error(self):
        """Test some other return from keystone.

        If we got anything else, something is wrong on the keystone
        side. We don't want to fail on our side.

        """
        self.mock_adap.get.return_value = fake_requests.FakeResponse(
            500, content="Oh noes!")
        self.assertTrue(identity.verify_project_id(mock.MagicMock(), "foo"))
        self.validate_common()

    def test_early_fail(self):
        """Test if we get a keystoneauth exception.

        If we get a random keystoneauth exception, fall back and
        assume the project exists.

        """
        self.mock_adap.get.side_effect = kse.ConnectionError()
        self.assertTrue(identity.verify_project_id(mock.MagicMock(), "foo"))
        self.validate_common()

    def test_wrong_version(self):
        """Test endpoint not found.

        EndpointNotFound will be made when the keystone v3 API is not
        found in the service catalog, or if the v2.0 endpoint was
        registered as the root endpoint. We treat this the same as 404.

        """
        self.mock_adap.get.side_effect = kse.EndpointNotFound()
        self.assertRaises(webob.exc.HTTPBadRequest,
                          identity.verify_project_id,
                          mock.MagicMock(), "foo")
        self.validate_common()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import fixtures

from oslo_utils.fixture import uuidsentinel

from nova import context as nova_context
from nova import exception
from nova import objects
from nova.tests.functional.libvirt import base
from nova.virt.libvirt import machine_type_utils


class LibvirtMachineTypeTest(base.ServersTestBase):

    microversion = 'latest'

    def setUp(self):
        super().setUp()
        self.context = nova_context.get_admin_context()

        # Add the q35 image to the glance fixture
        hw_machine_type_q35_image = copy.deepcopy(self.glance.image1)
        hw_machine_type_q35_image['id'] = uuidsentinel.q35_image_id
        hw_machine_type_q35_image['properties']['hw_machine_type'] = 'q35'
        self.glance.create(self.context, hw_machine_type_q35_image)

        # Create a pass-through mock around _get_guest_config to capture the
        # config of an instance so we can assert things about it later.
        # TODO(lyarwood): This seems like a useful thing to do in the libvirt
        # func tests for all computes we start?
        self.start_compute()
        self.guest_configs = {}
        orig_get_config = self.computes['compute1'].driver._get_guest_config

        def _get_guest_config(_self, *args, **kwargs):
            guest_config = orig_get_config(*args, **kwargs)
            instance = args[0]
            self.guest_configs[instance.uuid] = guest_config
            return self.guest_configs[instance.uuid]

        self.useFixture(fixtures.MonkeyPatch(
            'nova.virt.libvirt.LibvirtDriver._get_guest_config',
            _get_guest_config))

    def _create_servers(self):
        server_with = self._create_server(
            image_uuid=uuidsentinel.q35_image_id,
            networks='none',
        )
        server_without = self._create_server(
            image_uuid=self.glance.image1['id'],
            networks='none',
        )
        return (server_with, server_without)

    def _assert_machine_type(self, server_id, expected_machine_type):
        instance = objects.Instance.get_by_uuid(self.context, server_id)
        self.assertEqual(
            expected_machine_type,
            instance.image_meta.properties.hw_machine_type
        )
        self.assertEqual(
            expected_machine_type,
            instance.system_metadata['image_hw_machine_type']
        )
        self.assertEqual(
            expected_machine_type,
            self.guest_configs[server_id].os_mach_type
        )

    def _unset_machine_type(self, server_id):
        instance = objects.Instance.get_by_uuid(
            self.context,
            server_id,
        )
        instance.system_metadata.pop('image_hw_machine_type')
        instance.save()

    def test_init_host_register_machine_type(self):
        """Assert that the machine type of an instance is recorded during
           init_host if not already captured by an image prop.
        """
        self.flags(hw_machine_type='x86_64=pc', group='libvirt')

        server_with, server_without = self._create_servers()
        self._assert_machine_type(server_with['id'], 'q35')
        self._assert_machine_type(server_without['id'], 'pc')

        # Stop n-cpu and clear the recorded machine type from server_without to
        # allow init_host to register the machine type.
        self.computes['compute1'].stop()
        self._unset_machine_type(server_without['id'])

        self.flags(hw_machine_type='x86_64=pc-q35-1.2.3', group='libvirt')

        # Restart the compute
        self.computes['compute1'].start()

        # Assert the server_with remains pinned to q35
        self._assert_machine_type(server_with['id'], 'q35')

        # reboot the server so the config is rebuilt and _assert_machine_type
        # is able to pass. This just keeps the tests clean.
        self._reboot_server(server_without, hard=True)

        # Assert server_without now has a machine type of pc-q35-1.2.3 picked
        # up from [libvirt]hw_machine_type during init_host
        self._assert_machine_type(server_without['id'], 'pc-q35-1.2.3')

    def test_machine_type_after_config_change(self):
        """Assert new instances pick up a new default machine type after the
           config has been updated.
        """
        self.flags(hw_machine_type='x86_64=pc', group='libvirt')

        server_with, server_without = self._create_servers()
        self._assert_machine_type(server_with['id'], 'q35')
        self._assert_machine_type(server_without['id'], 'pc')

        self.flags(hw_machine_type='x86_64=pc-q35-1.2.3', group='libvirt')

        server_with_new, server_without_new = self._create_servers()
        self._assert_machine_type(server_with_new['id'], 'q35')
        self._assert_machine_type(server_without_new['id'], 'pc-q35-1.2.3')

    def test_machine_type_after_server_rebuild(self):
        """Assert that the machine type of an instance changes with a full
           rebuild of the instance pointing at a new image.
        """
        self.flags(hw_machine_type='x86_64=pc', group='libvirt')

        server_with, server_without = self._create_servers()
        self._assert_machine_type(server_with['id'], 'q35')
        self._assert_machine_type(server_without['id'], 'pc')

        # rebuild each server with the opposite image
        self._rebuild_server(
            server_with,
            '155d900f-4e14-4e4c-a73d-069cbf4541e6',
        )
        self._rebuild_server(
            server_without,
            uuidsentinel.q35_image_id
        )

        # Assert that the machine types were updated during the rebuild
        self._assert_machine_type(server_with['id'], 'pc')
        self._assert_machine_type(server_without['id'], 'q35')

    def _test_machine_type_after_server_reboot(self, hard=False):
        """Assert that the recorded machine types don't change with the
           reboot of a server, even when the underlying config changes.
        """
        self.flags(hw_machine_type='x86_64=pc', group='libvirt')

        server_with, server_without = self._create_servers()
        self._assert_machine_type(server_with['id'], 'q35')
        self._assert_machine_type(server_without['id'], 'pc')

        self.flags(hw_machine_type='x86_64=pc-q35-1.2.3', group='libvirt')

        self._reboot_server(server_with, hard=hard)
        self._reboot_server(server_without, hard=hard)

        # Assert that the machine types don't change after a reboot
        self._assert_machine_type(server_with['id'], 'q35')
        self._assert_machine_type(server_without['id'], 'pc')

    def test_machine_type_after_server_soft_reboot(self):
        self._test_machine_type_after_server_reboot()

    def test_machine_type_after_server_hard_reboot(self):
        self._test_machine_type_after_server_reboot(hard=True)

    def test_machine_type_get(self):
        self.flags(hw_machine_type='x86_64=pc', group='libvirt')

        server_with, server_without = self._create_servers()
        self.assertEqual(
            'q35',
            machine_type_utils.get_machine_type(
                self.context, server_with['id']
            )
        )
        self.assertEqual(
            'pc',
            machine_type_utils.get_machine_type(
                self.context, server_without['id']
            )
        )

    def test_machine_type_update_stopped(self):
        self.flags(hw_machine_type='x86_64=pc-1.2.3', group='libvirt')

        server = self._create_server(networks='none')
        self._assert_machine_type(server['id'], 'pc-1.2.3')

        self._stop_server(server)
        machine_type_utils.update_machine_type(
            self.context,
            server['id'],
            'pc-1.2.4'
        )

        self._start_server(server)
        self._assert_machine_type(server['id'], 'pc-1.2.4')

    def test_machine_type_update_blocked_active(self):
        self.flags(hw_machine_type='x86_64=pc-1.2.3', group='libvirt')

        server = self._create_server(networks='none')
        self._assert_machine_type(server['id'], 'pc-1.2.3')

        self.assertRaises(
            exception.InstanceInvalidState,
            machine_type_utils.update_machine_type,
            self.context,
            server['id'],
            'pc-1.2.4'
        )

    def test_machine_type_update_blocked_between_alias_and_versioned(self):
        self.flags(hw_machine_type='x86_64=pc', group='libvirt')

        server = self._create_server(networks='none')
        self._assert_machine_type(server['id'], 'pc')
        self._stop_server(server)

        self.assertRaises(
            exception.InvalidMachineTypeUpdate,
            machine_type_utils.update_machine_type,
            self.context,
            server['id'],
            'pc-1.2.4'
        )

    def test_machine_type_update_blocked_between_versioned_and_alias(self):
        self.flags(hw_machine_type='x86_64=pc-1.2.3', group='libvirt')

        server = self._create_server(networks='none')
        self._assert_machine_type(server['id'], 'pc-1.2.3')
        self._stop_server(server)

        self.assertRaises(
            exception.InvalidMachineTypeUpdate,
            machine_type_utils.update_machine_type,
            self.context,
            server['id'],
            'pc'
        )

    def test_machine_type_update_blocked_between_types(self):
        self.flags(hw_machine_type='x86_64=pc', group='libvirt')

        server = self._create_server(networks='none')
        self._assert_machine_type(server['id'], 'pc')
        self._stop_server(server)

        self.assertRaises(
            exception.InvalidMachineTypeUpdate,
            machine_type_utils.update_machine_type,
            self.context,
            server['id'],
            'q35'
        )

    def test_machine_type_update_force(self):
        self.flags(hw_machine_type='x86_64=pc', group='libvirt')

        server = self._create_server(networks='none')
        self._assert_machine_type(server['id'], 'pc')

        # Force through the update on an ACTIVE instance
        machine_type_utils.update_machine_type(
            self.context,
            server['id'],
            'q35',
            force=True
        )

        # Reboot the server so the config is updated so we can assert
        self._reboot_server(server, hard=True)
        self._assert_machine_type(server['id'], 'q35')

    def test_machine_type_list_unset_machine_type(self):
        self.flags(hw_machine_type='x86_64=pc', group='libvirt')

        server_with, server_without = self._create_servers()
        self._unset_machine_type(server_without['id'])

        instances = machine_type_utils.get_instances_without_type(self.context)
        self.assertEqual(
            server_without['id'],
            instances[0].uuid
        )

    def test_hw_machine_type_config_update_workflow(self):
        """Mimic a real world [libvirt]hw_machine_type config update workflow

        This test ties together various bits tested above into a complete
        workflow that should be typical of an operator attempting to update
        the host [libvirt]hw_machine_type configurable within their env.
        """
        self.flags(hw_machine_type='x86_64=pc', group='libvirt')

        # Launch some instances and assert their machine type
        server_with, server_without = self._create_servers()
        self._assert_machine_type(server_with['id'], 'q35')
        self._assert_machine_type(server_without['id'], 'pc')

        # Launch and shelve some additional instances
        server_shelved_with, server_shelved_without = self._create_servers()
        self._shelve_server(server_shelved_with)
        self._shelve_server(server_shelved_without)

        # Unset the machine type of the _without instances to mimic an existing
        # instance without a machine type set
        self._unset_machine_type(server_without['id'])
        self._unset_machine_type(server_shelved_without['id'])

        # Assert that both _without instances are listed by
        # get_instances_without_type as used by `nova-manage machine_type
        # list_unset` and `nova-status upgrade check`
        instances = machine_type_utils.get_instances_without_type(self.context)
        instance_uuids = [i.uuid for i in instances]
        self.assertIn(
            server_without['id'],
            instance_uuids
        )
        self.assertIn(
            server_shelved_without['id'],
            instance_uuids
        )

        # Restart the compute service
        self.computes['compute1'].stop()
        self.computes['compute1'].start()

        # Assert that after a service restart only server_shelved_without
        # remains listed by get_instances_without_type as init_host has
        # populated a machine type for server_without
        instances = machine_type_utils.get_instances_without_type(self.context)
        self.assertEqual(
            server_shelved_without['id'],
            instances[0].uuid
        )

        # Manually update the machine type of server_shelved_without using
        # machine_type_utils as used by `nova-manage machine_type update`
        machine_type_utils.update_machine_type(
            self.context,
            server_shelved_without['id'],
            'pc',
        )

        # Assert that no instances are listed so we can go ahead and change the
        # [libvirt]hw_machine_type config on the compute
        self.assertFalse(
            machine_type_utils.get_instances_without_type(self.context)
        )

        # Change the actual config on the compute
        self.flags(hw_machine_type='x86_64=pc-q35-1.2', group='libvirt')

        # Assert the existing instances remain the same after being rebooted or
        # unshelved, rebuilding their domain configs
        self._reboot_server(server_with, hard=True)
        self._reboot_server(server_without, hard=True)
        self._assert_machine_type(server_with['id'], 'q35')
        self._assert_machine_type(server_without['id'], 'pc')

        self._unshelve_server(server_shelved_with)
        self._unshelve_server(server_shelved_without)
        self._assert_machine_type(server_shelved_with['id'], 'q35')
        self._assert_machine_type(server_shelved_without['id'], 'pc')

        # Assert that new instances are spawned with the expected machine types
        server_with_new, server_without_new = self._create_servers()
        self._assert_machine_type(server_with_new['id'], 'q35')
        self._assert_machine_type(server_without_new['id'], 'pc-q35-1.2')



# Copyright 2016 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_policy import policy

from nova.policies import base


BASE_POLICY_NAME = 'os_compute_api:os-extended-server-attributes'


extended_server_attributes_policies = [
    policy.DocumentedRuleDefault(
        name=BASE_POLICY_NAME,
        check_str=base.PROJECT_ADMIN,
        description="""Return extended attributes for server.

This rule will control the visibility for a set of servers attributes:

- ``OS-EXT-SRV-ATTR:host``
- ``OS-EXT-SRV-ATTR:instance_name``
- ``OS-EXT-SRV-ATTR:reservation_id`` (since microversion 2.3)
- ``OS-EXT-SRV-ATTR:launch_index`` (since microversion 2.3)
- ``OS-EXT-SRV-ATTR:hostname`` (since microversion 2.3)
- ``OS-EXT-SRV-ATTR:kernel_id`` (since microversion 2.3)
- ``OS-EXT-SRV-ATTR:ramdisk_id`` (since microversion 2.3)
- ``OS-EXT-SRV-ATTR:root_device_name`` (since microversion 2.3)
- ``OS-EXT-SRV-ATTR:user_data`` (since microversion 2.3)

Microvision 2.75 added the above attributes in the ``PUT /servers/{server_id}``
and ``POST /servers/{server_id}/action (rebuild)`` API responses which are
also controlled by this policy rule, like the ``GET /servers*`` APIs.

Microversion 2.90 made the ``OS-EXT-SRV-ATTR:hostname`` attribute available to
all users, so this policy has no effect on that field for microversions 2.90
and greater. Controlling the visibility of this attribute for all microversions
is therefore deprecated and will be removed in a future release.
""",
        operations=[
            {
                'method': 'GET',
                'path': '/servers/{id}'
            },
            {
                'method': 'GET',
                'path': '/servers/detail'
            },
            {
                'method': 'PUT',
                'path': '/servers/{server_id}'
            },
            {
                'method': 'POST',
                'path': '/servers/{server_id}/action (rebuild)'
            }
        ],
        scope_types=['project']
    ),
]


def list_rules():
    return extended_server_attributes_policies



#    Copyright 2015 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

from oslo_log import log
from oslo_serialization import jsonutils
from oslo_utils import versionutils

from nova import exception
from nova.objects import base as obj_base
from nova.objects import fields

LOG = log.getLogger(__name__)
OS_VIF_DELEGATION = 'os_vif_delegation'


@obj_base.NovaObjectRegistry.register
class VIFMigrateData(obj_base.NovaObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    # The majority of the fields here represent a port binding on the
    # **destination** host during a live migration. The vif_type, among
    # other fields, could be different from the existing binding on the
    # source host, which is represented by the "source_vif" field.
    fields = {
        'port_id': fields.StringField(),
        'vnic_type': fields.StringField(),  # TODO(sean-k-mooney): make enum?
        'vif_type': fields.StringField(),
        # vif_details is a dict whose contents are dependent on the vif_type
        # and can be any number of types for the values, so we just store it
        # as a serialized dict
        'vif_details_json': fields.StringField(),
        # profile is in the same random dict of terrible boat as vif_details
        # so it's stored as a serialized json string
        'profile_json': fields.StringField(),
        'host': fields.StringField(),
        # The source_vif attribute is a copy of the VIF network model
        # representation of the port on the source host which can be used
        # for filling in blanks about the VIF (port) when building a
        # configuration reference for the destination host.
        # NOTE(mriedem): This might not be sufficient based on how the
        # destination host is configured for all vif types. See the note in
        # the libvirt driver here: https://review.opendev.org/#/c/551370/
        # 29/nova/virt/libvirt/driver.py@7036
        'source_vif': fields.Field(fields.NetworkVIFModel()),
    }

    @property
    def vif_details(self):
        if 'vif_details_json' not in self:
            return {}
        return jsonutils.loads(self.vif_details_json)

    @vif_details.setter
    def vif_details(self, vif_details_dict):
        self.vif_details_json = jsonutils.dumps(vif_details_dict)

    @property
    def profile(self):
        if 'profile_json' not in self:
            return {}
        return jsonutils.loads(self.profile_json)

    @profile.setter
    def profile(self, profile_dict):
        self.profile_json = jsonutils.dumps(profile_dict)

    @property
    def supports_os_vif_delegation(self):
        return self.profile.get(OS_VIF_DELEGATION, False)

    # TODO(stephenfin): add a proper delegation field instead of storing this
    # info in the profile catch-all blob
    @supports_os_vif_delegation.setter
    def supports_os_vif_delegation(self, supported):
        # we can't simply set the attribute using dict notation since the
        # getter returns a copy of the data, not the data itself
        self.profile = dict(
            self.profile or {}, **{OS_VIF_DELEGATION: supported})

    def get_dest_vif(self):
        """Get a destination VIF representation of this object.

        This method takes the source_vif and updates it to include the
        destination host port binding information using the other fields
        on this object.

        :return: nova.network.model.VIF object
        """
        if 'source_vif' not in self:
            raise exception.ObjectActionError(
                action='get_dest_vif', reason='source_vif is not set')
        vif = copy.deepcopy(self.source_vif)
        vif['type'] = self.vif_type
        vif['vnic_type'] = self.vnic_type
        vif['profile'] = self.profile
        vif['details'] = self.vif_details
        vif['delegate_create'] = self.supports_os_vif_delegation
        return vif

    @classmethod
    def create_skeleton_migrate_vifs(cls, vifs):
        """Create migrate vifs for live migration.

        :param vifs: a list of VIFs.
        :return: list of VIFMigrateData object corresponding to the provided
                 VIFs.
        """
        vif_mig_data = []

        for vif in vifs:
            mig_vif = cls(port_id=vif['id'], source_vif=vif)
            vif_mig_data.append(mig_vif)
        return vif_mig_data


@obj_base.NovaObjectRegistry.register
class LibvirtLiveMigrateNUMAInfo(obj_base.NovaObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        # NOTE(artom) We need a 1:many cardinality here, so DictOfIntegers with
        # its 1:1 cardinality cannot work here. cpu_pins can have a single
        # guest CPU pinned to multiple host CPUs.
        'cpu_pins': fields.DictOfSetOfIntegersField(),
        # NOTE(artom) Currently we never pin a guest cell to more than a single
        # host cell, so cell_pins could be a DictOfIntegers, but
        # DictOfSetOfIntegers is more future-proof.
        'cell_pins': fields.DictOfSetOfIntegersField(),
        'emulator_pins': fields.SetOfIntegersField(),
        'sched_vcpus': fields.SetOfIntegersField(),
        'sched_priority': fields.IntegerField(),
    }


@obj_base.NovaObjectRegistry.register_if(False)
class LiveMigrateData(obj_base.NovaObject):
    # Version 1.0: Initial version
    # Version 1.1: Added old_vol_attachment_ids field.
    # Version 1.2: Added wait_for_vif_plugged
    # Version 1.3: Added vifs field.
    VERSION = '1.3'

    fields = {
        'is_volume_backed': fields.BooleanField(),
        'migration': fields.ObjectField('Migration'),
        # old_vol_attachment_ids is a dict used to store the old attachment_ids
        # for each volume so they can be restored on a migration rollback. The
        # key is the volume_id, and the value is the attachment_id.
        # TODO(mdbooth): This field was made redundant by change Ibe9215c0. We
        # should eventually remove it.
        'old_vol_attachment_ids': fields.DictOfStringsField(),
        # wait_for_vif_plugged is set in pre_live_migration on the destination
        # compute host based on the [compute]/live_migration_wait_for_vif_plug
        # config option value; a default value is not set here since the
        # default for the config option may change in the future
        'wait_for_vif_plugged': fields.BooleanField(),
        'vifs': fields.ListOfObjectsField('VIFMigrateData'),
    }


@obj_base.NovaObjectRegistry.register
class LibvirtLiveMigrateBDMInfo(obj_base.NovaObject):
    # VERSION 1.0 : Initial version
    # VERSION 1.1 : Added encryption_secret_uuid for tracking volume secret
    #               uuid created on dest during migration with encrypted vols.
    VERSION = '1.1'

    fields = {
        # FIXME(danms): some of these can be enums?
        'serial': fields.StringField(),
        'bus': fields.StringField(),
        'dev': fields.StringField(),
        'type': fields.StringField(),
        'format': fields.StringField(nullable=True),
        'boot_index': fields.IntegerField(nullable=True),
        'connection_info_json': fields.StringField(),
        'encryption_secret_uuid': fields.UUIDField(nullable=True),
    }

    def obj_make_compatible(self, primitive, target_version):
        super(LibvirtLiveMigrateBDMInfo, self).obj_make_compatible(
            primitive, target_version)
        target_version = versionutils.convert_version_to_tuple(target_version)
        if target_version < (1, 1) and 'encryption_secret_uuid' in primitive:
            del primitive['encryption_secret_uuid']

    # NOTE(danms): We don't have a connection_info object right
    # now, and instead mostly store/pass it as JSON that we're
    # careful with. When we get a connection_info object in the
    # future, we should use it here, so make this easy to convert
    # for later.
    @property
    def connection_info(self):
        return jsonutils.loads(self.connection_info_json)

    @connection_info.setter
    def connection_info(self, info):
        self.connection_info_json = jsonutils.dumps(info)

    def as_disk_info(self):
        info_dict = {
            'dev': self.dev,
            'bus': self.bus,
            'type': self.type,
        }
        if self.obj_attr_is_set('format') and self.format:
            info_dict['format'] = self.format
        if self.obj_attr_is_set('boot_index') and self.boot_index is not None:
            info_dict['boot_index'] = str(self.boot_index)
        return info_dict


@obj_base.NovaObjectRegistry.register
class LibvirtLiveMigrateData(LiveMigrateData):
    # Version 1.0: Initial version
    # Version 1.1: Added target_connect_addr
    # Version 1.2: Added 'serial_listen_ports' to allow live migration with
    #              serial console.
    # Version 1.3: Added 'supported_perf_events'
    # Version 1.4: Added old_vol_attachment_ids
    # Version 1.5: Added src_supports_native_luks
    # Version 1.6: Added wait_for_vif_plugged
    # Version 1.7: Added dst_wants_file_backed_memory
    # Version 1.8: Added file_backed_memory_discard
    # Version 1.9: Inherited vifs from LiveMigrateData
    # Version 1.10: Added dst_numa_info, src_supports_numa_live_migration, and
    #               dst_supports_numa_live_migration fields
    VERSION = '1.10'

    fields = {
        'filename': fields.StringField(),
        # FIXME: image_type should be enum?
        'image_type': fields.StringField(),
        'block_migration': fields.BooleanField(),
        'disk_over_commit': fields.BooleanField(),
        'disk_available_mb': fields.IntegerField(nullable=True),
        'is_shared_instance_path': fields.BooleanField(),
        'is_shared_block_storage': fields.BooleanField(),
        'instance_relative_path': fields.StringField(),
        'graphics_listen_addr_vnc': fields.IPAddressField(nullable=True),
        'graphics_listen_addr_spice': fields.IPAddressField(nullable=True),
        'serial_listen_addr': fields.StringField(nullable=True),
        'serial_listen_ports': fields.ListOfIntegersField(),
        'bdms': fields.ListOfObjectsField('LibvirtLiveMigrateBDMInfo'),
        'target_connect_addr': fields.StringField(nullable=True),
        'supported_perf_events': fields.ListOfStringsField(),
        # TODO(lyarwood): No longer used, drop in version 2.0
        'src_supports_native_luks': fields.BooleanField(),
        'dst_wants_file_backed_memory': fields.BooleanField(),
        # TODO(lyarwood): No longer used, drop in version 2.0
        'file_backed_memory_discard': fields.BooleanField(),
        # TODO(artom) (src|dst)_supports_numa_live_migration are only used as
        # flags to indicate that the compute host is new enough to perform a
        # NUMA-aware live migration. Remove in version 2.0.
        'src_supports_numa_live_migration': fields.BooleanField(),
        'dst_supports_numa_live_migration': fields.BooleanField(),
        'dst_numa_info': fields.ObjectField('LibvirtLiveMigrateNUMAInfo'),
    }

    def obj_make_compatible(self, primitive, target_version):
        super(LibvirtLiveMigrateData, self).obj_make_compatible(
            primitive, target_version)
        target_version = versionutils.convert_version_to_tuple(target_version)
        if (target_version < (1, 10) and
                'src_supports_numa_live_migration' in primitive):
            del primitive['src_supports_numa_live_migration']
        if (target_version < (1, 10) and
                'dst_supports_numa_live_migration' in primitive):
            del primitive['dst_supports_numa_live_migration']
        if target_version < (1, 10) and 'dst_numa_info' in primitive:
            del primitive['dst_numa_info']
        if target_version < (1, 9) and 'vifs' in primitive:
            del primitive['vifs']
        if target_version < (1, 8):
            if 'file_backed_memory_discard' in primitive:
                del primitive['file_backed_memory_discard']
        if target_version < (1, 7):
            if 'dst_wants_file_backed_memory' in primitive:
                del primitive['dst_wants_file_backed_memory']
        if target_version < (1, 6) and 'wait_for_vif_plugged' in primitive:
            del primitive['wait_for_vif_plugged']
        if target_version < (1, 5):
            if 'src_supports_native_luks' in primitive:
                del primitive['src_supports_native_luks']
        if target_version < (1, 4):
            if 'old_vol_attachment_ids' in primitive:
                del primitive['old_vol_attachment_ids']
        if target_version < (1, 3):
            if 'supported_perf_events' in primitive:
                del primitive['supported_perf_events']
        if target_version < (1, 2):
            if 'serial_listen_ports' in primitive:
                del primitive['serial_listen_ports']
        if target_version < (1, 1) and 'target_connect_addr' in primitive:
            del primitive['target_connect_addr']

    def is_on_shared_storage(self):
        return self.is_shared_block_storage or self.is_shared_instance_path


@obj_base.NovaObjectRegistry.register
class HyperVLiveMigrateData(LiveMigrateData):
    # Version 1.0: Initial version
    # Version 1.1: Added is_shared_instance_path
    # Version 1.2: Added old_vol_attachment_ids
    # Version 1.3: Added wait_for_vif_plugged
    # Version 1.4: Inherited vifs from LiveMigrateData
    VERSION = '1.4'

    fields = {'is_shared_instance_path': fields.BooleanField()}

    def obj_make_compatible(self, primitive, target_version):
        super(HyperVLiveMigrateData, self).obj_make_compatible(
            primitive, target_version)
        target_version = versionutils.convert_version_to_tuple(target_version)
        if target_version < (1, 4) and 'vifs' in primitive:
            del primitive['vifs']
        if target_version < (1, 3) and 'wait_for_vif_plugged' in primitive:
            del primitive['wait_for_vif_plugged']
        if target_version < (1, 2):
            if 'old_vol_attachment_ids' in primitive:
                del primitive['old_vol_attachment_ids']
        if target_version < (1, 1):
            if 'is_shared_instance_path' in primitive:
                del primitive['is_shared_instance_path']


@obj_base.NovaObjectRegistry.register
class PowerVMLiveMigrateData(LiveMigrateData):
    # Version 1.0: Initial version
    # Version 1.1: Added the Virtual Ethernet Adapter VLAN mappings.
    # Version 1.2: Added old_vol_attachment_ids
    # Version 1.3: Added wait_for_vif_plugged
    # Version 1.4: Inherited vifs from LiveMigrateData
    VERSION = '1.4'

    fields = {
        'host_mig_data': fields.DictOfNullableStringsField(),
        'dest_ip': fields.StringField(),
        'dest_user_id': fields.StringField(),
        'dest_sys_name': fields.StringField(),
        'public_key': fields.StringField(),
        'dest_proc_compat': fields.StringField(),
        'vol_data': fields.DictOfNullableStringsField(),
        'vea_vlan_mappings': fields.DictOfNullableStringsField(),
    }

    def obj_make_compatible(self, primitive, target_version):
        super(PowerVMLiveMigrateData, self).obj_make_compatible(
            primitive, target_version)
        target_version = versionutils.convert_version_to_tuple(target_version)
        if target_version < (1, 4) and 'vifs' in primitive:
            del primitive['vifs']
        if target_version < (1, 3) and 'wait_for_vif_plugged' in primitive:
            del primitive['wait_for_vif_plugged']
        if target_version < (1, 2):
            if 'old_vol_attachment_ids' in primitive:
                del primitive['old_vol_attachment_ids']
        if target_version < (1, 1):
            if 'vea_vlan_mappings' in primitive:
                del primitive['vea_vlan_mappings']


@obj_base.NovaObjectRegistry.register
class VMwareLiveMigrateData(LiveMigrateData):
    VERSION = '1.0'

    fields = {
        'cluster_name': fields.StringField(nullable=False),
        'datastore_regex': fields.StringField(nullable=False),
    }



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from nova.tests.unit.virt.libvirt.volume import test_volume
from nova.virt.libvirt.volume import lightos

from os_brick import initiator


class LibvirtLightVolumeDriverTestCase(test_volume.LibvirtVolumeBaseTestCase):

    @mock.patch('nova.utils.get_root_helper')
    @mock.patch('os_brick.initiator.connector.InitiatorConnector.factory')
    def test_libvirt_lightos_driver(self, mock_factory, mock_helper):
        mock_helper.return_value = 'sudo'
        lightos.LibvirtLightOSVolumeDriver(self.fake_host)
        mock_factory.assert_called_once_with(
            initiator.LIGHTOS, root_helper='sudo',
            device_scan_attempts=5)

    @mock.patch('os_brick.initiator.connector.InitiatorConnector.factory',
        new=mock.Mock())
    def test_libvirt_lightos_driver_connect(self):
        lightos_driver = lightos.LibvirtLightOSVolumeDriver(
            self.fake_host)
        config = {'server_ip': '127.0.0.1', 'server_port': 9898}
        disk_info = {
            'id': '1234567',
            'name': 'aLightVolume',
            'conf': config}
        connection_info = {'data': disk_info}
        lightos_driver.connector.connect_volume.return_value = (
            {'path': '/dev/dms1234567'})

        lightos_driver.connect_volume(connection_info, None)

        lightos_driver.connector.connect_volume.assert_called_once_with(
            connection_info['data'])
        self.assertEqual(
            '/dev/dms1234567',
            connection_info['data']['device_path'])

    @mock.patch('os_brick.initiator.connector.InitiatorConnector.factory',
        new=mock.Mock(return_value=mock.Mock()))
    def test_libvirt_lightos_driver_disconnect(self):
        lightos_driver = lightos.LibvirtLightOSVolumeDriver(self.connr)
        disk_info = {
            'path': '/dev/dms1234567', 'name': 'aLightosVolume',
                    'type': 'raw', 'dev': 'vda1', 'bus': 'pci0',
                    'device_path': '/dev/dms123456'}
        connection_info = {'data': disk_info}
        lightos_driver.disconnect_volume(connection_info, None)
        lightos_driver.connector.disconnect_volume.assert_called_once_with(
            disk_info, None, force=False)

        # Verify force=True
        lightos_driver.connector.disconnect_volume.reset_mock()
        lightos_driver.disconnect_volume(connection_info, None, force=True)
        lightos_driver.connector.disconnect_volume.assert_called_once_with(
            disk_info, None, force=True)

    @mock.patch('os_brick.initiator.connector.InitiatorConnector.factory',
        new=mock.Mock(return_value=mock.Mock()))
    def test_libvirt_lightos_driver_get_config(self):
        lightos_driver = lightos.LibvirtLightOSVolumeDriver(self.fake_host)
        device_path = '/dev/fake-dev'
        connection_info = {'data': {'device_path': device_path}}

        conf = lightos_driver.get_config(connection_info, self.disk_info)
        tree = conf.format_dom()

        self.assertEqual('block', tree.get('type'))
        self.assertEqual(device_path, tree.find('./source').get('dev'))
        self.assertEqual('raw', tree.find('./driver').get('type'))



# Copyright (c) 2010 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_config import types

vnc_group = cfg.OptGroup(
    'vnc',
    title='VNC options',
    help="""
Virtual Network Computer (VNC) can be used to provide remote desktop
console access to instances for tenants and/or administrators.""")

ALL_OPTS = [
    cfg.BoolOpt(
        'enabled',
        default=True,
        deprecated_group='DEFAULT',
        deprecated_name='vnc_enabled',
        help="""
Enable VNC related features.

Guests will get created with graphical devices to support this. Clients
(for example Horizon) can then establish a VNC connection to the guest.
"""),

    cfg.HostAddressOpt(
        'server_listen',
        default='127.0.0.1',
        help="""
The IP address or hostname on which an instance should listen to for
incoming VNC connection requests on this node.
"""),

    cfg.HostAddressOpt(
        'server_proxyclient_address',
        default='127.0.0.1',
        help="""
Private, internal IP address or hostname of VNC console proxy.

The VNC proxy is an OpenStack component that enables compute service
users to access their instances through VNC clients.

This option sets the private address to which proxy clients, such as
``nova-novncproxy``, should connect to.
"""),

    cfg.URIOpt(
        'novncproxy_base_url',
        default='http://127.0.0.1:6080/vnc_auto.html',
        deprecated_group='DEFAULT',
        help="""
Public address of noVNC VNC console proxy.

The VNC proxy is an OpenStack component that enables compute service
users to access their instances through VNC clients. noVNC provides
VNC support through a websocket-based client.

This option sets the public base URL to which client systems will
connect. noVNC clients can use this address to connect to the noVNC
instance and, by extension, the VNC sessions.

If using noVNC >= 1.0.0, you should use ``vnc_lite.html`` instead of
``vnc_auto.html``.

Related options:

* novncproxy_host
* novncproxy_port
"""),

]

CLI_OPTS = [
    cfg.StrOpt(
        'novncproxy_host',
        default='0.0.0.0',
        deprecated_group='DEFAULT',
        help="""
IP address that the noVNC console proxy should bind to.

The VNC proxy is an OpenStack component that enables compute service
users to access their instances through VNC clients. noVNC provides
VNC support through a websocket-based client.

This option sets the private address to which the noVNC console proxy
service should bind to.

Related options:

* novncproxy_port
* novncproxy_base_url
"""),

    cfg.PortOpt(
        'novncproxy_port',
        default=6080,
        deprecated_group='DEFAULT',
        help="""
Port that the noVNC console proxy should bind to.

The VNC proxy is an OpenStack component that enables compute service
users to access their instances through VNC clients. noVNC provides
VNC support through a websocket-based client.

This option sets the private port to which the noVNC console proxy
service should bind to.

Related options:

* novncproxy_host
* novncproxy_base_url
"""),
    cfg.ListOpt(
        'auth_schemes',
        item_type=types.String(choices=(
            ('none', 'Allow connection without authentication'),
            ('vencrypt', 'Use VeNCrypt authentication scheme'),
        )),
        default=['none'],
        help="""
The authentication schemes to use with the compute node.

Control what RFB authentication schemes are permitted for connections between
the proxy and the compute host. If multiple schemes are enabled, the first
matching scheme will be used, thus the strongest schemes should be listed
first.

Related options:

* ``[vnc]vencrypt_client_key``, ``[vnc]vencrypt_client_cert``: must also be set
"""),
    cfg.StrOpt(
        'vencrypt_client_key',
        help="""The path to the client certificate PEM file (for x509)

The fully qualified path to a PEM file containing the private key which the VNC
proxy server presents to the compute node during VNC authentication.

Related options:

* ``vnc.auth_schemes``: must include ``vencrypt``
* ``vnc.vencrypt_client_cert``: must also be set
"""),
    cfg.StrOpt(
        'vencrypt_client_cert',
        help="""The path to the client key file (for x509)

The fully qualified path to a PEM file containing the x509 certificate which
the VNC proxy server presents to the compute node during VNC authentication.

Realted options:

* ``vnc.auth_schemes``: must include ``vencrypt``
* ``vnc.vencrypt_client_key``: must also be set
"""),
    cfg.StrOpt(
        'vencrypt_ca_certs',
        help="""The path to the CA certificate PEM file

The fully qualified path to a PEM file containing one or more x509 certificates
for the certificate authorities used by the compute node VNC server.

Related options:

* ``vnc.auth_schemes``: must include ``vencrypt``
"""),
]

ALL_OPTS.extend(CLI_OPTS)


def register_opts(conf):
    conf.register_group(vnc_group)
    conf.register_opts(ALL_OPTS, group=vnc_group)


def register_cli_opts(conf):
    conf.register_cli_opts(CLI_OPTS, group=vnc_group)


def list_opts():
    return {vnc_group: ALL_OPTS}
